/*
 * Copyright (c) 2002  Paul Vojta and the Xdvik project team
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * PAUL VOJTA OR ANY OTHER AUTHOR OF THIS SOFTWARE BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

/*
  SU: adopted from xdvi-22.50, popups.c

  This implements a print popup (and a `save to file' popup).  My idea
  was to use the same window layout framework for both tasks, since
  there's a big overlap in the widgets needed.  However after looking
  at the mess I've created, I'm no longer convinced this was a good
  idea. (I reckon it's slightly better than the code duplication when
  doing it in separate files, but only *slightly* ...)
  
  We need a more intelligent approach that bans this ugly kind of
  `data proliferation' (those global arrays of widgets and helper
  structures).  Custom Widgets? Factory methods? Needs investigation.
 */

#include <string.h>
#include "xdvi-config.h"
#include "xdvi.h"
#include "my-vsnprintf.h"
#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>

#include "print-dialog.h"
#include "print-log.h"
#include "events.h"
#include "dvi-init.h"
#include "string-utils.h"
#include "util.h"
#include "x_util.h"
#include "message-window.h"
#include "pagesel.h"
#include "sfSelFile.h"
#include "dvisel.h" /* for select_marked_pages() */
#include "xlwradio.h"
#include "statusline.h"

/* Xlib and Xutil are already included */


#include <X11/Xatom.h>
#include <X11/StringDefs.h>

#ifdef MOTIF
# include <Xm/BulletinB.h>
# include <Xm/DialogS.h>
# include <Xm/PanedW.h>
# include <Xm/MessageB.h>
# include <Xm/LabelG.h>
# include <Xm/Form.h>
# include <Xm/Frame.h>
# include <Xm/ToggleBG.h>
# include <Xm/Text.h>
# include <Xm/TextF.h>
# include <Xm/PushB.h>
# include <Xm/Protocols.h>
#else /* MOTIF */
# include <X11/Shell.h>
# include <X11/Xaw/Paned.h>
# include <X11/Xaw/Box.h>
# include <X11/Xaw/Form.h>
# include <X11/Xaw/Label.h>
# include <X11/Xaw/Command.h>
# include <X11/Xaw/Toggle.h>
# include <X11/Xaw/AsciiText.h>
#endif /* MOTIF */

#ifndef	MAX
# define MAX(i, j)       ( (i) > (j) ? (i) : (j) )
#endif

#if HAVE_VFORK_H
# include <vfork.h>
#endif

#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WIFEXITED
# define WIFEXITED(status)	(((status) & 255) == 0)
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(status)	((unsigned)(status) >> 8)
#endif
#ifndef WIFSIGNALED
# ifndef WIFSTOPPED
#  define WIFSTOPPED(status)	(((status) & 0xff) == 0x7f)
# endif
# define WIFSIGNALED(status)	(!WIFSTOPPED(status) && !WIFEXITED(status))
#endif
#ifndef WTERMSIG
# define WTERMSIG(status)	((status) & 0x7f)
#endif

/* if POSIX O_NONBLOCK is not available, use O_NDELAY */
#if !defined(O_NONBLOCK) && defined(O_NDELAY)
# define O_NONBLOCK O_NDELAY
#endif

#ifdef EWOULDBLOCK
# ifdef EAGAIN
#  define AGAIN_CONDITION	(errno == EWOULDBLOCK || errno == EAGAIN)
# else
#  define AGAIN_CONDITION	(errno == EWOULDBLOCK)
# endif
#else /* EWOULDBLOCK */
# ifdef EAGAIN
#  define AGAIN_CONDITION	(errno == EAGAIN)
# endif
#endif /* EWOULDBLOCK */

#if HAVE_POLL
# include <poll.h>
# define XIO_IN POLLIN
# define XIO_OUT POLLOUT
#else
# define XIO_IN 1
# define XIO_OUT 2
#endif /* HAVE_POLL */


/*
  global widgets, and arrays of global widgets (mosly, one
  for the printing and one for the saving dialog). Needs
  cleanup!
*/

/* top-level widgets: one for printing, one for saving to file */
static Widget print_shell = NULL;
static Widget save_shell = NULL;

#ifdef MOTIF
static Widget from_label[FILE_ALL] = { None, None };
#endif /* MOTIF */

static Widget to_printer_radio;
static Widget to_file_radio_or_label[FILE_ALL] = { None, None };
static Widget to_label[FILE_ALL] = { None, None };
static Widget of_label[FILE_ALL] = { None, None };
static Widget range_all_radio[FILE_ALL] = { None, None };
static Widget range_marked_radio[FILE_ALL] = { None, None };
static Widget range_from_to_radio[FILE_ALL] = { None, None };
static Widget cancel_button[FILE_ALL] = { None, None };
static Widget ok_button[FILE_ALL] = { None, None };
static Widget print_to_printer_text;
static Widget print_to_file_text[FILE_ALL] = { None, None };
static Widget print_to_file_button[FILE_ALL] = { None, None };
static Widget dvips_options_text;
static Widget page_from_text[FILE_ALL] = { None, None };
static Widget page_to_text[FILE_ALL] = { None, None };

static int print_active = 0;	/* -1 if window is unmanaged, 0 if not created yet, 1 if managed */
static int save_active = 0;	/* -1 if window is unmanaged, 0 if not created yet, 1 if managed */

typedef enum print_radio_t_ {
    NO_PRINT_VAL = -1,
    PRINT_TO_PRINTER = 1,
    PRINT_TO_FILE,
    SAVE_TO_FILE
} print_radio_t;

typedef enum page_radio_t_ {
    NO_PAGE_VAL = -1,
    PRINT_ALL = 1,
    PRINT_MARKED,
    PRINT_RANGE,
    SAVE_ALL,
    SAVE_MARKED,
    SAVE_RANGE
} page_radio_t;

static Boolean is_print_button(page_radio_t flag) {
    return (flag <= PRINT_RANGE);
}


typedef struct dialog_info_t_ {
    char *printer_name;
    char *file_name;
    char *expanded_filename;	/* fully canonicalized filename, not visible in window, but used in error messages */
    char *dvips_options;
    char *from_page;
    char *to_page;
    print_radio_t print_radio;
    page_radio_t range_radio;
} dialog_info_t;

/* global structures for saving print info between invocations */
static dialog_info_t save_pr_info = { NULL, NULL, NULL, NULL, NULL, NULL, NO_PRINT_VAL, NO_PAGE_VAL };
static dialog_info_t save_save_info = { NULL, NULL, NULL, NULL, NULL, NULL, NO_PRINT_VAL, NO_PAGE_VAL };

/*
  these contain the currently selected values, or defaults;
  used in communication between callbacks
*/
static dialog_info_t curr_pr_info = { NULL, NULL, NULL, NULL, NULL, NULL, PRINT_TO_PRINTER, PRINT_ALL };
static dialog_info_t curr_save_info = { NULL, NULL, NULL, NULL, NULL, NULL, SAVE_TO_FILE, SAVE_ALL };

static void print_precheck(print_or_save_action_t act);

#define PRINT_LOG_TIMEOUT 1500 /* leave log window open for 1.5 seconds */

/*
  ============================================================
  generic utility functions
  ============================================================
*/

static void
warn_no_marked_pages(void)
{
    XBell(DISP, 0);
    do_popup_message(MSG_ERR,
		     "This option can only be selected "
		     "when there are some pages marked in the page list. "
                    "You can mark pages by clicking an entry in the \"Mark\" menu, "
                    "by using the \"m\" key on the keyboard, "
		     "or by clicking/dragging Mouse-2 in the page list.",
		     "There are currently no pages marked!");
}

/*
  Return the printer name, eventually followed by `-P$PRINTER',
  since the `-o' option of dvips doesn't evaluate the environment
  variable.
*/
static char *
get_printer_name(void)
{
    char *printcmd, *printer;

    if (resource.dvips_printer_str != NULL) {
	return xstrdup(resource.dvips_printer_str);
    }
    printcmd = xstrdup("lpr");
    printer = getenv("PRINTER");
    if (printer != NULL) {
	printcmd = xstrcat(printcmd, " -P");
	printcmd = xstrcat(printcmd, printer);
    }
    return printcmd;
}

/* return basename of `old_name' with `old_ext' replaced by `new_ext', in fresh memory */
static char *
get_new_file_name(const char *old_name, const char *old_ext, const char *new_ext)
{
    /* old_name contains the normalized DVI file name, with path and extension */
    char *basename, *new_name;
    size_t len;
    
    basename = strrchr(old_name, DIR_SEPARATOR);
		       
    if (basename != NULL) {
	basename++;
	new_name = xmalloc(strlen(basename) + strlen(new_ext) + 1);
	new_name = strcpy(new_name, basename);
    }
    else {
	new_name = xmalloc(strlen(old_name) + strlen(new_ext) + 1);
	new_name = strcpy(new_name, old_name);
    }

    len = strlen(new_name);
    strcpy(new_name + len - strlen(old_ext), new_ext);
    return new_name;
}

/* get initial values for dialog texts */
static void
get_initial_dialog_values(dialog_info_t *info, print_or_save_action_t act)
{
    if (act == FILE_PRINT) {
	if (info->printer_name == NULL)
	    info->printer_name = get_printer_name();
    }
    
    if (info->file_name == NULL) { /* FIXME: what to do if dvi_name has changed since last invocation? */
	if (act == FILE_PRINT)
	    info->file_name = get_new_file_name(dvi_name, ".dvi", ".ps");
	else
	    info->file_name = get_new_file_name(dvi_name, ".dvi", "_pp.dvi");
    }

    if (act == FILE_PRINT) {
	if (info->dvips_options == NULL) {
	    if (resource.dvips_options_str == NULL)
		info->dvips_options = xstrdup("");
	    else
		info->dvips_options = xstrdup(resource.dvips_options_str);
	}
    }

    if (info->from_page == NULL) {
	info->from_page = xmalloc(length_of_int(current_page + pageno_correct) + 1);
	sprintf(info->from_page, "%d", current_page + pageno_correct);
    }

    if (info->to_page == NULL) {
	info->to_page = xmalloc(length_of_int(current_page + pageno_correct) + 1);
	sprintf(info->to_page, "%d", current_page + pageno_correct);
    }
}

/* adjust two widgets vertically */
void
adjust_vertically(Widget w1, Widget w2, int default_dist)
{
    Dimension h1, h2;
#if MOTIF
    XtVaGetValues(w1, XmNheight, &h1, NULL);
    XtVaGetValues(w2, XmNheight, &h2, NULL);
    XtVaSetValues(w1, XmNtopOffset, default_dist + (h2 - h1) / 2, NULL);
/*     XtVaSetValues(w2, XmNtopOffset, default_dist + (h2 - h1) / 2, NULL); */
#else
    XtVaGetValues(w1, XtNheight, &h1, NULL);
    XtVaGetValues(w2, XtNheight, &h2, NULL);
    XtVaSetValues(w1, XtNvertDistance, default_dist + (h2 - h1) / 2, NULL);
    XtVaSetValues(w2, XtNvertDistance, default_dist + (h2 - h1) / 2, NULL);
#endif /* MOTIF */
}


#if MOTIF
static void
set_sensitivity(int i)
{
    Boolean sensitivity;
    
    if (i == curr_pr_info.print_radio) /* if no change */
	return;

    curr_pr_info.print_radio = i;

    sensitivity = (i == PRINT_TO_PRINTER);
    XtSetSensitive(print_to_printer_text, sensitivity);

    sensitivity ^= (True ^ False);
    XtSetSensitive(print_to_file_text[FILE_PRINT], sensitivity);
    XtSetSensitive(print_to_file_button[FILE_PRINT], sensitivity);
}
#endif /* MOTIF */

/* callbacks */
static void
cb_print_cancel(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(client_data);
    UNUSED(call_data);
    
    XtPopdown(print_shell);
    print_active = False;
}

static void
cb_save_cancel(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(client_data);
    UNUSED(call_data);
    
    XtPopdown(save_shell);
    save_active = False;
}

static void
cb_print(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(call_data);

    if (client_data != NULL) {
	print_or_save_action_t act = (print_or_save_action_t)client_data;
	print_precheck(act);
    }
    else {
	print_precheck(FILE_PRINT);
    }
}

#ifndef MOTIF
static void
xaw_print_cancel(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);
    cb_print_cancel(w, NULL, NULL);
}
    
static void
xaw_save_cancel(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);
    cb_save_cancel(w, NULL, NULL);
}

static void xaw_print_act_go(Widget, XEvent *, String *, Cardinal *);
static void xaw_save_act_go(Widget, XEvent *, String *, Cardinal *);
#endif

#ifndef MOTIF
static XtActionsRec popup_actions[] = {
    {"saveInternal",		xaw_save_act_go },
    {"printInternal",		xaw_print_act_go},
    {"WM_print_cancel",		xaw_print_cancel },
    {"WM_save_cancel",		xaw_save_cancel  },
};

/*
  enable the `print to' button if user presses key or mouse button
  inside that input field
*/
static void
xaw_print_handle_key(Widget widget, XtPointer closure, XEvent *ev, Boolean *cont)
{
    char *number;
    int i;

    number = (char *)closure;
    i = atoi(number);
    ASSERT(i == 2 || i == 3, "Button index should be 2 or 3");
    UNUSED(widget);
    UNUSED(ev);
    UNUSED(cont);

    if (i == 2) {
	XawToggleSetCurrent(to_printer_radio, (XtPointer)PRINT_TO_PRINTER);
    }
    else {
	ASSERT(to_file_radio_or_label[FILE_PRINT] != NULL, "");
	XawToggleSetCurrent(to_file_radio_or_label[FILE_PRINT], (XtPointer)PRINT_TO_FILE);
    }
}

/*
  enable the `print range' button if user presses key or mouse button
  inside that input field
*/
static void
xaw_range_handle_key(Widget widget, XtPointer closure, XEvent *ev, Boolean *cont)
{
    print_or_save_action_t act = (print_or_save_action_t)closure;
    UNUSED(widget);
    UNUSED(ev);
    UNUSED(cont);
    
    if (act == FILE_PRINT && curr_pr_info.range_radio != PRINT_RANGE) {
	ASSERT(range_all_radio[act] != NULL, "button should have been initialized");
	XawToggleSetCurrent(range_all_radio[act], (XtPointer)PRINT_RANGE);
    }
    else if (act == FILE_SAVE && curr_save_info.range_radio != SAVE_RANGE) {
	ASSERT(range_all_radio[act] != NULL, "button should have been initialized");
	XawToggleSetCurrent(range_all_radio[act], (XtPointer)SAVE_RANGE);
    }
}


static void
xaw_print_act_go(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    print_precheck(FILE_PRINT);
}

static void
xaw_save_act_go(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
    UNUSED(w);
    UNUSED(event);
    UNUSED(params);
    UNUSED(num_params);

    print_precheck(FILE_SAVE);
}

#else /* MOTIF */

static void
xm_range_set_sensitivity(page_radio_t flag)
{
    Boolean sensitivity;
    print_or_save_action_t act;
    if (is_print_button(flag)) {
	sensitivity = (curr_pr_info.range_radio == PRINT_RANGE);
	act = FILE_PRINT;
    }
    else {
	sensitivity = (curr_save_info.range_radio == SAVE_RANGE);
	act = FILE_SAVE;
    }

    XtSetSensitive(from_label[act], sensitivity);
    XtSetSensitive(page_from_text[act], sensitivity);
    XtSetSensitive(to_label[act], sensitivity);
    XtSetSensitive(page_to_text[act], sensitivity);
    XtSetSensitive(of_label[act], sensitivity);
}

#endif /* not MOTIF */

static void
cb_print_vs_file(Widget w, XtPointer client_data, XtPointer call_data)
{
#ifdef MOTIF
    if (((XmToggleButtonCallbackStruct *) call_data)->set) {
	if (w != to_printer_radio) {
	    XmToggleButtonGadgetSetState(to_printer_radio, False, False);
	    to_printer_radio = w;
	    set_sensitivity((int) client_data);
	    XmProcessTraversal(curr_pr_info.print_radio == PRINT_TO_PRINTER
			       ? print_to_printer_text
			       : print_to_file_text[FILE_PRINT],
			       XmTRAVERSE_CURRENT);
	}
    }
    else if (w == to_printer_radio)
	XmToggleButtonGadgetSetState(w, True, False);
#else /* MOTIF */
    int i = (int) XawToggleGetCurrent(to_printer_radio);

    UNUSED(w);
    UNUSED(client_data);
    UNUSED(call_data);

    /*
      NOTE: Unlike non-k xdvi, we don't use set_sensitivity for the
      Xaw version, since the text inside the textfield cannot be made
      insensitive, so it won't be obvious to users why they cannot
      change the text.  Also, the page range isn't set insensitive
      either.
     */
    if (i != 0)
	curr_pr_info.print_radio = i;
#endif /* MOTIF */
}



static void
cb_range(Widget w, XtPointer client_data, XtPointer call_data)
{
    page_radio_t flag = (page_radio_t)client_data;
    print_or_save_action_t act;
#ifdef MOTIF
    XmToggleButtonCallbackStruct *s = (XmToggleButtonCallbackStruct *)call_data;
    if ((flag == PRINT_MARKED || flag == SAVE_MARKED) && !pageinfo_have_marked_pages()) {
	if (s->set == 0) { /* return if value hasn't changed; this is kind of a hack,
			      otherwise popup will be displayed twice */
	    return;
	}

	/* tried to invoke `print marked' without any pages being marked? */
	warn_no_marked_pages();
	/* fix highlighting of button */
	if (flag == PRINT_MARKED) {
	    XmToggleButtonGadgetSetState(range_marked_radio[FILE_PRINT], False, False);
	    if (curr_pr_info.range_radio == PRINT_ALL) {
		XmToggleButtonGadgetSetState(range_all_radio[FILE_PRINT], True, False);
		XmProcessTraversal(range_all_radio[FILE_PRINT], XmTRAVERSE_CURRENT);
	    }
	    else {
		XmToggleButtonGadgetSetState(range_from_to_radio[FILE_PRINT], True, False);
		XmProcessTraversal(range_from_to_radio[FILE_PRINT], XmTRAVERSE_CURRENT);
	    }
	}
	else {
	    XmToggleButtonGadgetSetState(range_marked_radio[FILE_SAVE], False, False);
	    if (curr_save_info.range_radio == SAVE_ALL) {
		XmToggleButtonGadgetSetState(range_all_radio[FILE_SAVE], True, False);
		XmProcessTraversal(range_all_radio[FILE_SAVE], XmTRAVERSE_CURRENT);
	    }
	    else {
		XmToggleButtonGadgetSetState(range_from_to_radio[FILE_SAVE], True, False);
		XmProcessTraversal(range_from_to_radio[FILE_SAVE], XmTRAVERSE_CURRENT);
	    }
	}
	/* 	curr_pr_info.range_radio = (int)client_data; */
	return;
    }

    if (is_print_button(flag)) {
	curr_pr_info.range_radio = (int)client_data;
	act = FILE_PRINT;
    }
    else {
	curr_save_info.range_radio = (int)client_data;
	act = FILE_SAVE;
    }
    
    XmToggleButtonGadgetSetState(range_all_radio[act], False, False);
    XmToggleButtonGadgetSetState(range_marked_radio[act], False, False);
    XmToggleButtonGadgetSetState(range_from_to_radio[act], False, False);
    XmToggleButtonGadgetSetState(w, True, False);
    
    xm_range_set_sensitivity(flag);
#else /* MOTIF */
    int i;
    UNUSED(call_data);
    if (is_print_button(flag)) {
	act = FILE_PRINT;
    }
    else {
	act = FILE_SAVE;
    }

    UNUSED(w);

    i = (int) XawToggleGetCurrent(range_all_radio[act]);

    if (i != 0) {
	if ((i == PRINT_MARKED || i == SAVE_MARKED) && !pageinfo_have_marked_pages()) {
	    /* tried to invoke `print marked' without any pages being marked? */
	    warn_no_marked_pages();
	    if (act == FILE_PRINT)
		XawToggleSetCurrent(range_all_radio[act], (XtPointer)curr_pr_info.range_radio);
	    else
		XawToggleSetCurrent(range_all_radio[act], (XtPointer)curr_save_info.range_radio);
	    return;
	}
	if (act == FILE_PRINT)
	    curr_pr_info.range_radio = i;
	else
	    curr_save_info.range_radio = i;
    }
#endif /* MOTIF */
}


static void
cb_browse(Widget w, XtPointer client_data, XtPointer call_data)
{
    char *fname;
    print_or_save_action_t act = (print_or_save_action_t)client_data;

    UNUSED(call_data);
#ifdef MOTIF
    UNUSED(w);
#else
    if (act == FILE_PRINT)
	xaw_print_handle_key(w, (XtPointer)"3", NULL, NULL);
#endif
    
    if (act == FILE_SAVE)
	fname = XsraSelFile(top_level, "Xdvi: Save to file", "Save to file:",
			    "Ok", "Cancel",
			    NULL, NULL,
			    "*.dvi", False);
    else
	fname = XsraSelFile(top_level, "Xdvi: Print to file", "Print to file:",
			    "Ok", "Cancel",
			    NULL, NULL,
			    "*.ps", False);
    
    if (fname != NULL) {
	XtVaSetValues(print_to_file_text[act],
#ifdef MOTIF
		      XmNvalue, fname, XmNcursorPosition, strlen(fname),
#else
		      XtNstring, fname,
#endif
		      NULL);
    }
    free(fname);
}

/* create a popup dialog for printing a DVI file OR saving pages from a DVI file */
static Widget
create_print_or_save_window(print_or_save_action_t act, char *ofstring)
{
    Widget top_level_shell;
    Atom WM_DELETE_WINDOW;
#ifndef MOTIF
    Widget form, paned, box;
    Widget f01form, f02form; /* dummy forms to get indentation consistent */
    Widget r1form;
    Widget r1label;
    Widget r2form, r3form, r4form = 0;
    Widget r4label;
    Widget r5label;
    Widget r6form;
    Widget r61form;
    Widget r7form;

    XtTranslations xlats, xlats2, wm_translations;
    XtAccelerators accels2;
    int ddist;

    /* handy abbrevs */
#define HORIZONTAL_RESIZING_NO	XtNleft, XtChainLeft, XtNright, XtChainLeft
#define HORIZONTAL_RESIZING_YES XtNleft, XtChainLeft, XtNright, XtChainRight
#define VERTICAL_RESIZING_NO	XtNtop, XtChainTop, XtNbottom, XtChainTop

    XtAddActions(popup_actions, XtNumber(popup_actions));

    if (act == FILE_SAVE) {
	wm_translations = XtParseTranslationTable("<Message>WM_PROTOCOLS: WM_save_cancel()\n"
						  "<Key>q: WM_save_cancel()");
    }
    else {
	wm_translations = XtParseTranslationTable("<Message>WM_PROTOCOLS: WM_print_cancel()\n"
						  "<Key>q: WM_print_cancel()");
    }

    top_level_shell = XtVaCreatePopupShell(act == FILE_SAVE ? "save" : "print",
					   transientShellWidgetClass,
					   top_level,
					   XtNtitle, act == FILE_SAVE ? "Save dvi file" : "Print dvi file",
					   XtNmappedWhenManaged, False,
					   XtNtransientFor, top_level,
					   XtNallowShellResize, True,
					   XtNvisual, our_visual,
					   XtNcolormap, our_colormap,
					   XtNtranslations, wm_translations,
					   NULL);

    WM_DELETE_WINDOW = XInternAtom(XtDisplay(top_level_shell), "WM_DELETE_WINDOW", False);

    paned = XtVaCreateManagedWidget("paned", panedWidgetClass, top_level_shell, NULL);
	
    form = XtVaCreateManagedWidget("form", formWidgetClass,
				   paned,
				   XtNallowResize, True,
				   NULL);
    XtVaGetValues(form, XtNdefaultDistance, &ddist, NULL);

    xlats = XtParseTranslationTable("<EnterWindow>:highlight(Always)\n"
				    "<LeaveWindow>:unhighlight()\n"
				    "<Btn1Down>,<Btn1Up>:set()notify()");

    if (act == FILE_PRINT)
	xlats2 = XtParseTranslationTable("<Key>Return:printInternal()");
    else
	xlats2 = XtParseTranslationTable("<Key>Return:saveInternal()");


    if (act == FILE_PRINT) { /* selection `to printer/to file' */
	/* FIXME: set real sizes here */
	Dimension w_avg = 220, w_curr;
	f01form = XtVaCreateManagedWidget("dummy_print_to_label_form", formWidgetClass,
					  form,
					  XtNresizable, True,
					  XtNborderWidth, 0,
/* 					  XtNvertDistance, 10, */
					  HORIZONTAL_RESIZING_YES,
					  NULL);
	r1label = XtVaCreateManagedWidget("print_to", labelWidgetClass,
					  f01form,
					  XtNlabel, "Print to: ",
					  XtNborderWidth, 0,
					  HORIZONTAL_RESIZING_NO,
					  NULL);
	
	r2form = XtVaCreateManagedWidget("print_to_printer_form", formWidgetClass,
					 form,
					 XtNresizable, True,
					 XtNborderWidth, 0,
					 XtNvertDistance, 0,
					 XtNfromVert, f01form,
					 HORIZONTAL_RESIZING_YES,
					 XtNallowResize, True,
					 NULL);
	
	accels2 = XtParseAcceleratorTable("<Btn1Down>,<Btn1Up>:set()notify()");
	
	to_printer_radio = XtVaCreateManagedWidget("to_printer",
#ifdef XAW
						   radioWidgetClass,
#else
						   toggleWidgetClass,
#endif
						   r2form,
						   XtNlabel, "Printer: ",
						   XtNborderWidth, 0,
						   XtNhighlightThickness, 1,
						   XtNradioData, PRINT_TO_PRINTER,
						   XtNstate, True, /* enable this button by default */
						   XtNtranslations, xlats,
						   XtNaccelerators, accels2,
						   HORIZONTAL_RESIZING_NO,
						   VERTICAL_RESIZING_NO,
						   NULL);
	
	XtAddCallback(to_printer_radio, XtNcallback, cb_print_vs_file, NULL);
	XtInstallAccelerators(r2form, to_printer_radio);
	
	print_to_printer_text = XtVaCreateManagedWidget("printer_text", asciiTextWidgetClass,
							r2form,
							XtNwidth, w_avg,
							XtNdataCompression, False,
							XtNeditType, XawtextEdit,
/*  							XtNresize, XawtextResizeWidth, */
							XtNfromHoriz, to_printer_radio,
							HORIZONTAL_RESIZING_YES,
							VERTICAL_RESIZING_NO,
/*  							XtNscrollHorizontal, XawtextScrollAlways, */
							NULL);
	XtOverrideTranslations(print_to_printer_text, xlats2);
	XtAddEventHandler(print_to_printer_text, KeyPressMask|ButtonPressMask, False,
			  xaw_print_handle_key, (XtPointer)"2");
	
	adjust_vertically(to_printer_radio, print_to_printer_text, ddist);
	
	r3form = XtVaCreateManagedWidget("print_to_file_form", formWidgetClass,
					 form,
					 XtNborderWidth, 0,
					 XtNvertDistance, 0,
					 XtNfromVert, r2form,
					 HORIZONTAL_RESIZING_YES,
					 NULL);

	to_file_radio_or_label[act] = XtVaCreateManagedWidget("to_file",
#ifdef XAW
							      radioWidgetClass,
#else
							      toggleWidgetClass,
#endif
							      r3form,
							      XtNhighlightThickness, 1,
							      XtNborderWidth, 0,
							      XtNlabel, "File: ",
							      XtNradioGroup, to_printer_radio,
							      XtNradioData, PRINT_TO_FILE,
							      XtNtranslations, xlats,
							      XtNaccelerators, accels2,
							      HORIZONTAL_RESIZING_NO,
							      VERTICAL_RESIZING_NO,
							      NULL);
	XtAddCallback(to_file_radio_or_label[act], XtNcallback, cb_print_vs_file, NULL);
	XtInstallAccelerators(r3form, to_file_radio_or_label[act]);

	print_to_file_text[act] = XtVaCreateManagedWidget("file_text", asciiTextWidgetClass,
							  r3form,
							  XtNwidth, w_avg,
							  XtNdataCompression, False,
							  XtNeditType, XawtextEdit,
							  XtNfromHoriz, to_file_radio_or_label[act],
							  HORIZONTAL_RESIZING_YES,
							  VERTICAL_RESIZING_NO,
							  NULL);
	XtOverrideTranslations(print_to_file_text[act], xlats2);
	XtAddEventHandler(print_to_file_text[act], KeyPressMask|ButtonPressMask, False,
			  xaw_print_handle_key, (XtPointer)"3");
#define DIST_1 8 /* FIXME */
	print_to_file_button[act] = XtVaCreateManagedWidget("file_button", commandWidgetClass,
							    r3form,
							    XtNlabel, "Browse ...",
							    XtNfromHoriz, print_to_file_text[act],
							    XtNhorizDistance, 8,
							    /* attach to right of form, no resizing: */
							    XtNleft, XtChainRight, XtNright, XtChainRight,
							    VERTICAL_RESIZING_NO,
							    NULL);

	XtVaGetValues(print_to_file_button[act], XtNwidth, &w_curr, NULL);
	XtVaSetValues(print_to_file_text[act], XtNwidth, w_avg - w_curr - DIST_1 - 2, NULL);
#undef DIST_1
	
	XtAddCallback(print_to_file_button[act], XtNcallback, cb_browse, (XtPointer)act);
/*  	adjust_vertically(to_file_radio_or_label[act], print_to_file_text[act], ddist + 5); */
/*  	adjust_vertically(to_file_radio_or_label[act], print_to_file_button[act], ddist + 5); */
/*  	adjust_vertically(print_to_file_button[act], to_file_radio_or_label[act], ddist + 5); */

	/* Row 4: other dvips options */

	r4form = XtVaCreateManagedWidget("dvips_options_form", formWidgetClass,
					 form,
					 XtNborderWidth, 0,
					 XtNfromVert, r3form,
					 XtNvertDistance, 0,
					 HORIZONTAL_RESIZING_YES,
					 NULL);

	
	r4label = XtVaCreateManagedWidget("dvips_options_name", labelWidgetClass,
					  r4form,
					  XtNlabel, "Dvips options:",
					  XtNborderWidth, 0,
					  HORIZONTAL_RESIZING_NO,
					  VERTICAL_RESIZING_NO,
					  NULL);

	dvips_options_text = XtVaCreateManagedWidget("dvips_options_text", asciiTextWidgetClass,
						     r4form,
						     XtNwidth, w_avg,
						     XtNdataCompression, False,
						     XtNeditType, XawtextEdit,
						     XtNfromHoriz, r4label,
						     HORIZONTAL_RESIZING_YES,
						     VERTICAL_RESIZING_NO,
						     NULL);
	XtOverrideTranslations(dvips_options_text, xlats2);

	{ /* align left edges of input fields */
	    Dimension w1, w2, w3, w_max;

	    XtVaGetValues(to_printer_radio, XtNwidth, &w1, NULL);
	    XtVaGetValues(to_file_radio_or_label[act], XtNwidth, &w2, NULL);
	    XtVaGetValues(r4label, XtNwidth, &w3, NULL);

	    w_max = MAX(MAX(w1, w2), w3);
	    
	    XtVaSetValues(print_to_printer_text, XtNhorizDistance, ddist + (w_max - w1), NULL);
	    XtVaSetValues(print_to_file_text[act], XtNhorizDistance, ddist + (w_max - w2), NULL);
	    XtVaSetValues(dvips_options_text, XtNhorizDistance, ddist + (w_max - w3), NULL);
	}
    }
    else { /* not printing, but saving */
	r1form = XtVaCreateManagedWidget("save_to_file_form", formWidgetClass,
					 form,
					 XtNborderWidth, 0,
					 HORIZONTAL_RESIZING_YES,
					 VERTICAL_RESIZING_NO,
					 NULL);
	to_file_radio_or_label[act] = XtVaCreateManagedWidget("to_file", labelWidgetClass,
							      r1form,
							      XtNlabel, "Save to file: ",
							      XtNborderWidth, 0,
							      HORIZONTAL_RESIZING_NO,
							      VERTICAL_RESIZING_NO,
							      NULL);
	print_to_file_text[act] = XtVaCreateManagedWidget("file_text", asciiTextWidgetClass,
							  r1form,
							  XtNdataCompression, False,
/* 							  XtNvertDistance, ddist + 5, */
							  XtNeditType, XawtextEdit,
							  XtNfromHoriz, to_file_radio_or_label[act],
							  HORIZONTAL_RESIZING_YES,
							  VERTICAL_RESIZING_NO,
							  NULL);
	print_to_file_button[act] = XtVaCreateManagedWidget("file_button", commandWidgetClass,
							    r1form,
							    XtNlabel, "Browse ...",
							    XtNfromHoriz, print_to_file_text[act],
/* 							    XtNvertDistance, ddist + 5, */
							    XtNhorizDistance, 5,
							    /* attach to right of form, no resizing: */
							    XtNleft, XtChainRight, XtNright, XtChainRight,
							    VERTICAL_RESIZING_NO,
							    NULL);
	XtAddCallback(print_to_file_button[act], XtNcallback, cb_browse, (XtPointer)act);
	XtOverrideTranslations(print_to_file_text[act], xlats2);
	
	
	adjust_vertically(to_file_radio_or_label[act], print_to_file_text[act], ddist + 5);
 	adjust_vertically(to_file_radio_or_label[act], print_to_file_button[act], ddist + 5);
/*  	adjust_vertically(print_to_file_text[act], to_file_radio_or_label[act], ddist + 5); */
 	adjust_vertically(print_to_file_button[act], to_file_radio_or_label[act], ddist + 5);
    }

    /* Rows 5-7:  page selection */

    f02form = XtVaCreateManagedWidget("dummy_pages_form", formWidgetClass,
				      form,
				      XtNfromVert, act == FILE_PRINT ? r4form : print_to_file_text[act],
				      XtNresizable, True,
				      XtNborderWidth, 0,
/* 				      XtNvertDistance, 10, */
				      HORIZONTAL_RESIZING_NO,
				      NULL);
    
    r5label = XtVaCreateManagedWidget("range_lab", labelWidgetClass,
				      f02form,
				      XtNlabel, "Pages:",
				      XtNborderWidth, 0,
				      HORIZONTAL_RESIZING_NO,
				      NULL);

    r6form = XtVaCreateManagedWidget("page_range_form", formWidgetClass,
				     form,
				     XtNborderWidth, 0,
				     XtNfromVert, f02form,
/* 				     XtNvertDistance, 0, */
				     HORIZONTAL_RESIZING_NO,
				     NULL);

    accels2 = XtParseAcceleratorTable("<Btn1Down>,<Btn1Up>:set()notify()");
    range_all_radio[act] = XtVaCreateManagedWidget("range_all",
#ifdef XAW
						   radioWidgetClass,
#else
						   toggleWidgetClass,
#endif
						   r6form,
						   XtNlabel, "All ",
						   XtNborderWidth, 0,
						   XtNhighlightThickness, 1,
						   XtNradioData, act == FILE_PRINT ? PRINT_ALL : SAVE_ALL,
						   XtNstate, True, /* enable this button by default */
						   XtNtranslations, xlats,
						   XtNaccelerators, accels2,
/* 						   XtNvertDistance, 0, */
						   HORIZONTAL_RESIZING_NO,
						   VERTICAL_RESIZING_NO,
						   NULL);
    if (act == FILE_PRINT)
	XtAddCallback(range_all_radio[act], XtNcallback, cb_range, (XtPointer)PRINT_ALL);
    else
	XtAddCallback(range_all_radio[act], XtNcallback, cb_range, (XtPointer)SAVE_ALL);
    XtInstallAccelerators(r6form, range_all_radio[act]);

    r61form = XtVaCreateManagedWidget("range_marked_form", formWidgetClass,
				      form,
				      XtNborderWidth, 0,
				      XtNfromVert, r6form,
/* 				      XtNvertDistance, 0, */
				      HORIZONTAL_RESIZING_NO,
				      NULL);

    range_marked_radio[act] = XtVaCreateManagedWidget("range_marked",
#ifdef XAW
						      radioWidgetClass,
#else
						      toggleWidgetClass,
#endif

						      r61form,
						      XtNlabel, "Marked ",
						      XtNborderWidth, 0,
						      XtNhighlightThickness, 1,
						      XtNradioGroup, range_all_radio[act],
						      XtNradioData, act == FILE_PRINT ? PRINT_MARKED : SAVE_MARKED,
						      XtNtranslations, xlats,
						      XtNaccelerators, accels2,
/* 						      XtNvertDistance, 0, */
						      HORIZONTAL_RESIZING_NO,
						      VERTICAL_RESIZING_NO,
						      NULL);
    if (act == FILE_PRINT)
	XtAddCallback(range_marked_radio[act], XtNcallback, cb_range, (XtPointer)PRINT_RANGE);
    else
	XtAddCallback(range_marked_radio[act], XtNcallback, cb_range, (XtPointer)SAVE_RANGE);
    XtInstallAccelerators(r61form, range_marked_radio[act]);

    r7form = XtVaCreateManagedWidget("range_from_to_form", formWidgetClass,
				     form,
				     XtNborderWidth, 0,
				     XtNfromVert, r61form,
/* 				     XtNvertDistance, 0, */
				     HORIZONTAL_RESIZING_NO,
				     NULL);

    range_from_to_radio[act] = XtVaCreateManagedWidget("range_from_to",
#ifdef XAW
						       radioWidgetClass,
#else
						       toggleWidgetClass,
#endif
						       r7form,
						       XtNlabel, "From: ",
						       XtNborderWidth, 0,
						       XtNhighlightThickness, 1,
						       XtNradioGroup, range_all_radio[act],
						       XtNradioData, act == FILE_PRINT ? PRINT_RANGE : SAVE_RANGE,
						       XtNtranslations, xlats,
						       XtNaccelerators, accels2,
/* 						       XtNvertDistance, 0, */
						       HORIZONTAL_RESIZING_NO,
						       VERTICAL_RESIZING_NO,
						       NULL);
    if (act == FILE_PRINT)
	XtAddCallback(range_from_to_radio[act], XtNcallback, cb_range, (XtPointer)PRINT_RANGE);
    else
	XtAddCallback(range_from_to_radio[act], XtNcallback, cb_range, (XtPointer)SAVE_RANGE);
    XtInstallAccelerators(r7form, range_from_to_radio[act]);

    page_from_text[act] = XtVaCreateManagedWidget("range_from", asciiTextWidgetClass,
						  r7form,
						  XtNdataCompression, False,
						  XtNeditType, XawtextEdit,
						  XtNwidth, 50,
						  XtNfromHoriz, range_from_to_radio[act],
						  HORIZONTAL_RESIZING_NO,
						  VERTICAL_RESIZING_NO,
						  NULL);
    XtOverrideTranslations(page_from_text[act], xlats2);
    XtAddEventHandler(page_from_text[act], KeyPressMask|ButtonPressMask, False,
		      xaw_range_handle_key, (XtPointer)act);

    to_label[act] = XtVaCreateManagedWidget("range_to_lab", labelWidgetClass,
					    r7form,
					    XtNlabel, "to: ",
					    XtNborderWidth, 0,
					    XtNfromHoriz, page_from_text[act],
					    HORIZONTAL_RESIZING_NO,
					    VERTICAL_RESIZING_NO,
					    NULL);

    page_to_text[act] = XtVaCreateManagedWidget("range_to", asciiTextWidgetClass,
						r7form,
						XtNdataCompression, False,
						XtNeditType, XawtextEdit,
						XtNwidth, 50,
						XtNfromHoriz, to_label[act],
						HORIZONTAL_RESIZING_NO,
						VERTICAL_RESIZING_NO,
						NULL);
    XtOverrideTranslations(page_to_text[act], xlats2);
    XtAddEventHandler(page_to_text[act], KeyPressMask|ButtonPressMask, False,
		      xaw_range_handle_key, (XtPointer)act);

    of_label[act] = XtVaCreateManagedWidget("range_of", labelWidgetClass,
					    r7form,
					    XtNlabel, ofstring,
					    XtNborderWidth, 0,
					    XtNresizable, True,
					    XtNfromHoriz, page_to_text[act],
					    HORIZONTAL_RESIZING_NO,
					    VERTICAL_RESIZING_NO,
					    NULL);

    adjust_vertically(range_from_to_radio[act], page_from_text[act], ddist);
    adjust_vertically(to_label[act], page_from_text[act], ddist);
    adjust_vertically(of_label[act], page_from_text[act], ddist);

    /* box for the OK/Cancel button */
    box = XtVaCreateManagedWidget("box", formWidgetClass,
				  paned,
				  /* resizing by user isn't needed */
				  XtNshowGrip, False,
				  XtNdefaultDistance, 6, /* some padding */
				  /* resizing the window shouldn't influence this box,
				   * but only the pane widget
				   */
				  XtNskipAdjust, True,
				  XtNaccelerators, accels_cr,
				  NULL);
	
    ok_button[act] = XtVaCreateManagedWidget("ok", commandWidgetClass,
					     box,
					     XtNlabel, act == FILE_PRINT ? "Print" : "Save",
					     XtNaccelerators, accels_cr,
					     XtNtop, XtChainTop,
					     XtNbottom, XtChainBottom,
					     HORIZONTAL_RESIZING_NO,
					     NULL);
    if (act == FILE_PRINT)
	XtAddCallback(ok_button[act], XtNcallback, cb_print, (XtPointer)FILE_PRINT);
    else
	XtAddCallback(ok_button[act], XtNcallback, cb_print, (XtPointer)FILE_SAVE);
    
    XtInstallAccelerators(form, ok_button[act]);
    if (act == FILE_PRINT)
	XtInstallAccelerators(print_to_printer_text, ok_button[act]);
    XtInstallAccelerators(print_to_file_text[act], ok_button[act]);

    cancel_button[act] = XtVaCreateManagedWidget("cancel", commandWidgetClass,
						 box,
						 XtNlabel, "Cancel",
						 XtNfromHoriz, ok_button[act],
						 XtNbottom, XtChainBottom,
						 XtNjustify, XtJustifyRight,
						 XtNleft, XtChainRight,
						 XtNright, XtChainRight,
						 NULL);
    if (act == FILE_PRINT)
	XtAddCallback(cancel_button[act], XtNcallback, cb_print_cancel, NULL);
    else
	XtAddCallback(cancel_button[act], XtNcallback, cb_save_cancel, NULL);

    XtManageChild(top_level_shell);
    center_window(top_level_shell, top_level);
    XtMapWidget(top_level_shell);
    XSetWMProtocols(XtDisplay(top_level_shell), XtWindow(top_level_shell), &WM_DELETE_WINDOW, 1);

    if (((act == FILE_PRINT && save_pr_info.range_radio < 0)
	 || (act == FILE_SAVE && save_save_info.range_radio < 0))
	&& pageinfo_have_marked_pages()) {
	XawToggleSetCurrent(range_all_radio[act],
			    act == FILE_PRINT ? (XtPointer)PRINT_MARKED : (XtPointer)SAVE_MARKED);
    }
#else /* not MOTIF */
    Widget form, pane, box;
    Widget frame1, f1label, f1child;
    Widget r4label;
    Widget frame2, f2label, f2child;
    XmString str;
    
    top_level_shell = XtVaCreatePopupShell(act == FILE_SAVE ? "save" : "print",
					   xmDialogShellWidgetClass, top_level,
					   XmNtitle, act == FILE_SAVE ? "Save dvi file" : "Print dvi file",
					   XmNallowShellResize, True,
					   XmNdeleteResponse, XmDO_NOTHING, /* we'll take care of that ourselves */
					   XmNmappedWhenManaged, False, /* so that we can center it first */
					   XmNvisual, our_visual,
					   XmNcolormap, our_colormap,
					   NULL);

    /* make the window manager destroy action just pop down the dialog */
    WM_DELETE_WINDOW = XmInternAtom(XtDisplay(top_level_shell), "WM_DELETE_WINDOW", False);
    if (act == FILE_SAVE)
	XmAddWMProtocolCallback(top_level_shell, WM_DELETE_WINDOW, cb_save_cancel, NULL);
    else
	XmAddWMProtocolCallback(top_level_shell, WM_DELETE_WINDOW, cb_print_cancel, NULL);


    pane = XtVaCreateWidget("print_pane", xmPanedWindowWidgetClass, top_level_shell,
			    /* make sashes invisible */
			    XmNsashWidth, 1,
			    XmNsashHeight, 1,
			    /* turn separator off, since it gives ugly spacing with the frames */
			    XmNseparatorOn, False,
			    NULL);

    
    form = XtVaCreateWidget("form", xmFormWidgetClass,
			    pane,
			    XmNhorizontalSpacing, DDIST_MAJOR,
			    XmNverticalSpacing, DDIST_MAJOR,
			    XmNautoUnmanage, False,
			    NULL);

    if (act == FILE_PRINT) {
	/*  First frame (rows 1-3):  print to printer or file */

	frame1 = XtVaCreateWidget("printtoframe", xmFrameWidgetClass, form,
				  XmNmarginWidth, DDIST,
				  XmNmarginHeight, DDIST,
				  XmNtopAttachment, XmATTACH_FORM,
				  XmNleftAttachment, XmATTACH_FORM,
				  XmNrightAttachment, XmATTACH_FORM,
				  NULL);

	str = XmStringCreateLocalized("Print To:");
	f1label = XtVaCreateManagedWidget("title", xmLabelGadgetClass,
					  frame1,
					  XmNchildType, XmFRAME_TITLE_CHILD,
					  XmNlabelString, str,
					  NULL);
	XmStringFree(str);

	f1child = XtVaCreateWidget("form", xmFormWidgetClass, frame1,
				   XmNhorizontalSpacing, DDIST,
				   XmNverticalSpacing, DDIST,
				   XmNtopAttachment, XmATTACH_FORM,
				   XmNleftAttachment, XmATTACH_FORM,
				   XmNrightAttachment, XmATTACH_FORM,
				   NULL);

	str = XmStringCreateLocalized("Printer:");
	to_printer_radio = XtVaCreateManagedWidget("to_printer",
						   xmToggleButtonGadgetClass, f1child,
						   XmNlabelString, str,
						   XmNindicatorType, XmONE_OF_MANY,
						   XmNset, True,
						   XmNtopAttachment, XmATTACH_FORM,
						   XmNtopOffset, 0,
						   XmNleftAttachment, XmATTACH_FORM,
						   XmNleftOffset, 0,
						   NULL);
	XmStringFree(str);
	XtAddCallback(to_printer_radio, XmNvalueChangedCallback, cb_print_vs_file, (XtPointer)1);


	print_to_printer_text = XtVaCreateManagedWidget("print_text", xmTextFieldWidgetClass,
							f1child,
							XmNtopAttachment, XmATTACH_FORM,
							XmNtopOffset, 0,
							XmNleftAttachment, XmATTACH_WIDGET,
							XmNrightAttachment, XmATTACH_FORM,
							XmNleftWidget, to_printer_radio,
							NULL);
	XtAddCallback(print_to_printer_text, XmNactivateCallback, cb_print, NULL);

	adjust_vertically(to_printer_radio, print_to_printer_text, 0);

	str = XmStringCreateLocalized("File:");
	to_file_radio_or_label[act] = XtVaCreateManagedWidget("to_file", xmToggleButtonGadgetClass,
							      f1child,
							      XmNlabelString, str,
							      XmNindicatorType, XmONE_OF_MANY,
							      XmNtopAttachment, XmATTACH_WIDGET,
							      XmNtopWidget, print_to_printer_text,
							      XmNleftAttachment, XmATTACH_FORM,
							      XmNleftOffset, 0,
							      NULL);
	XmStringFree(str);
	XtAddCallback(to_file_radio_or_label[act], XmNvalueChangedCallback, cb_print_vs_file, (XtPointer)2);

	str = XmStringCreateLocalized("Browse ...");
	print_to_file_button[act] = XtVaCreateManagedWidget("file_button", xmPushButtonWidgetClass,
							    f1child,
							    XmNlabelString, str,
							    XmNsensitive, False,
							    XmNtopAttachment, XmATTACH_WIDGET,
							    XmNtopWidget, print_to_printer_text,
							    XmNrightAttachment, XmATTACH_FORM,
							    NULL);
	XmStringFree(str);	
	print_to_file_text[act] = XtVaCreateManagedWidget("file_text", xmTextFieldWidgetClass,
							  f1child,
							  XmNsensitive, False,
							  XmNtopAttachment, XmATTACH_WIDGET,
							  XmNtopWidget, print_to_printer_text,
							  XmNrightAttachment, XmATTACH_WIDGET,
							  XmNrightWidget, print_to_file_button[act],
							  XmNleftAttachment, XmATTACH_WIDGET,
							  XmNleftWidget, to_file_radio_or_label[act],
							  NULL);
	XtAddCallback(print_to_file_text[act], XmNactivateCallback, cb_print, NULL);
	adjust_vertically(to_file_radio_or_label[act], print_to_file_text[act], DDIST);

	XtAddCallback(print_to_file_button[act], XmNactivateCallback, cb_browse, (XtPointer)act);
	adjust_vertically(print_to_file_button[act], print_to_file_text[act], DDIST);

	{ /* align left edges of text widgets */
	    Dimension w1, w2;

	    XtVaGetValues(to_printer_radio, XmNwidth, &w1, NULL);
	    XtVaGetValues(to_file_radio_or_label[act], XmNwidth, &w2, NULL);
	    
	    if (w1 > w2) {
		Dimension offset = DDIST;
		offset += w1 - w2;
		XtVaSetValues(print_to_file_text[act], XmNleftOffset, offset, NULL);
	    }
	    else if (w2 > w1) {
		Dimension offset = DDIST;
		offset += w2 - w1;
		XtVaSetValues(print_to_printer_text, XmNleftOffset, offset, NULL);
	    }
	}

	XtManageChild(f1child);
	XtManageChild(frame1);

	/* Row 4:  additional dvips options */

	str = XmStringCreateLocalized("Dvips options:");
	r4label = XtVaCreateManagedWidget("dvips_options_name", xmLabelGadgetClass,
					  form,
					  XmNlabelString, str,
					  XmNtopAttachment, XmATTACH_WIDGET,
					  XmNtopWidget, frame1,
					  XmNleftAttachment, XmATTACH_FORM,
					  NULL);
	XmStringFree(str);

	dvips_options_text = XtVaCreateManagedWidget("dvips_options_text", xmTextFieldWidgetClass,
						     form,
						     XmNtopAttachment, XmATTACH_WIDGET,
						     XmNtopWidget, frame1,
						     XmNleftAttachment, XmATTACH_WIDGET,
						     XmNrightAttachment, XmATTACH_FORM,
						     XmNleftWidget, r4label,
						     XmNcolumns, 20,
						     NULL);
	XtAddCallback(dvips_options_text, XmNactivateCallback, cb_print, (XtPointer)act);

	adjust_vertically(r4label, dvips_options_text, DDIST_MAJOR);
    } /* if (act == FILE_PRINT) */
    else {
	str = XmStringCreateLocalized("Save to file:");
	to_file_radio_or_label[act] = XtVaCreateManagedWidget("to_file", xmLabelGadgetClass,
							      form,
							      XmNtopAttachment, XmATTACH_FORM,
							      XmNtopOffset, DDIST_MAJOR,
							      XmNleftAttachment, XmATTACH_FORM,
							      /* 							      XmNchildType, XmFRAME_TITLE_CHILD, */
							      XmNlabelString, str,
							      NULL);
	XmStringFree(str);
	str = XmStringCreateLocalized("Browse ...");
	print_to_file_button[act] = XtVaCreateManagedWidget("file_button", xmPushButtonWidgetClass,
							    form,
							    XmNlabelString, str,
							    XmNtopAttachment, XmATTACH_FORM,
							    XmNrightAttachment, XmATTACH_FORM,
							    NULL);
	print_to_file_text[act] = XtVaCreateManagedWidget("file_text", xmTextFieldWidgetClass,
							  form,
							  XmNtopAttachment, XmATTACH_FORM,
							  XmNtopOffset, DDIST_MAJOR,
							  XmNleftAttachment, XmATTACH_WIDGET,
							  XmNleftWidget, to_file_radio_or_label[act],
							  XmNrightAttachment, XmATTACH_WIDGET,
							  XmNrightWidget, print_to_file_button[act],
							  NULL);
	
	adjust_vertically(to_file_radio_or_label[act], print_to_file_text[act], DDIST_MAJOR);
	XtAddCallback(print_to_file_button[act], XmNactivateCallback, cb_browse, (XtPointer)act);
	adjust_vertically(print_to_file_button[act], print_to_file_text[act], DDIST_MAJOR);
    }
    
    /* Second frame (rows 5-7):  page selection */
    frame2 = XtVaCreateWidget("print_to_frame", xmFrameWidgetClass, form,
			      XmNmarginWidth, DDIST,
			      XmNmarginHeight, DDIST,
			      XmNresizable, True,
			      XmNtopAttachment, XmATTACH_WIDGET,
			      XmNtopWidget, act == FILE_PRINT ? dvips_options_text : print_to_file_text[act],
			      XmNleftAttachment, XmATTACH_FORM,
			      XmNrightAttachment, XmATTACH_FORM,
			      NULL);

    str = XmStringCreateLocalized("Pages:");
    f2label = XtVaCreateManagedWidget("title", xmLabelGadgetClass,
				      frame2,
				      XmNchildType, XmFRAME_TITLE_CHILD,
				      XmNlabelString, str,
				      NULL);
    XmStringFree(str);

    f2child = XtVaCreateWidget("form", xmFormWidgetClass, frame2,
			       XmNhorizontalSpacing, DDIST,
			       XmNverticalSpacing, DDIST,
			       XmNresizable, True,
			       NULL);

    str = XmStringCreateLocalized("All");
    range_all_radio[act] = XtVaCreateManagedWidget("all",
						   xmToggleButtonGadgetClass, f2child,
						   XmNlabelString, str,
						   XmNindicatorType, XmONE_OF_MANY,
						   XmNset, True,
						   XmNtopAttachment, XmATTACH_FORM,
						   XmNtopOffset, 0,
						   XmNleftAttachment, XmATTACH_FORM,
						   XmNleftOffset, 0,
						   NULL);
    XmStringFree(str);
    if (act == FILE_PRINT)
	XtAddCallback(range_all_radio[act], XmNvalueChangedCallback, cb_range, (XtPointer)PRINT_ALL);
    else
	XtAddCallback(range_all_radio[act], XmNvalueChangedCallback, cb_range, (XtPointer)SAVE_ALL);

    str = XmStringCreateLocalized("Marked");
    range_marked_radio[act] =  XtVaCreateManagedWidget("marked",
						       xmToggleButtonGadgetClass, f2child,
						       XmNlabelString, str,
						       XmNindicatorType, XmONE_OF_MANY,
						       XmNtopAttachment, XmATTACH_WIDGET,
						       XmNtopWidget, range_all_radio[act],
						       XmNleftAttachment, XmATTACH_FORM,
						       XmNleftOffset, 0,
						       NULL);
    XmStringFree(str);
    /*
      Need to use both XmNarmCallback and XmNvalueChangedCallback. When only using
      XmNvalueChangedCallback, clicking the button another time after warning has
      shown would just work. Now, we ignore such events explicitly in cb_range,
      by checking the `set' field in the XmToggleButtonCallbackStruct *.
    */
    if (act == FILE_PRINT) {
	XtAddCallback(range_marked_radio[act], XmNvalueChangedCallback, cb_range, (XtPointer)PRINT_MARKED);
	XtAddCallback(range_marked_radio[act], XmNarmCallback, cb_range, (XtPointer)PRINT_MARKED);
    }
    else {
	XtAddCallback(range_marked_radio[act], XmNvalueChangedCallback, cb_range, (XtPointer)SAVE_MARKED);
	XtAddCallback(range_marked_radio[act], XmNarmCallback, cb_range, (XtPointer)SAVE_MARKED);
    }
	
    str = XmStringCreateLocalized("Range:");
    range_from_to_radio[act] = XtVaCreateManagedWidget("range",
						       xmToggleButtonGadgetClass, f2child,
						       XmNlabelString, str,
						       XmNindicatorType, XmONE_OF_MANY,
						       XmNtopAttachment, XmATTACH_WIDGET,
						       XmNtopWidget, range_marked_radio[act],
						       XmNleftAttachment, XmATTACH_FORM,
						       XmNleftOffset, 0,
						       NULL);
    XmStringFree(str);

    if (act == FILE_PRINT)
	XtAddCallback(range_from_to_radio[act], XmNvalueChangedCallback, cb_range, (XtPointer)PRINT_RANGE);
    else
	XtAddCallback(range_from_to_radio[act], XmNvalueChangedCallback, cb_range, (XtPointer)SAVE_RANGE);

    str = XmStringCreateLocalized("From");
    from_label[act] = XtVaCreateManagedWidget("from", xmLabelGadgetClass,
					      f2child,
					      XmNlabelString, str,
					      XmNsensitive, False,
					      XmNtopAttachment, XmATTACH_WIDGET,
					      XmNtopWidget, range_marked_radio[act],
					      XmNleftAttachment, XmATTACH_WIDGET,
					      XmNleftWidget, range_from_to_radio[act],
					      NULL);
    XmStringFree(str);

    page_from_text[act] = XtVaCreateManagedWidget("from_page",
						  xmTextFieldWidgetClass, f2child,
						  XmNcolumns, 5,
						  XmNsensitive, False,
						  XmNtopAttachment, XmATTACH_WIDGET,
						  XmNtopWidget, range_marked_radio[act],
						  XmNleftAttachment, XmATTACH_WIDGET,
						  XmNleftWidget, from_label[act],
						  NULL);
    XtAddCallback(page_from_text[act], XmNactivateCallback, cb_print, (XtPointer)act);

    str = XmStringCreateLocalized("to");
    to_label[act] = XtVaCreateManagedWidget("to", xmLabelGadgetClass,
					    f2child,
					    XmNlabelString, str,
					    XmNsensitive, False,
					    XmNtopAttachment, XmATTACH_WIDGET,
					    XmNtopWidget, range_marked_radio[act],
					    XmNleftAttachment, XmATTACH_WIDGET,
					    XmNleftWidget, page_from_text[act],
					    NULL);
    XmStringFree(str);

    page_to_text[act] = XtVaCreateManagedWidget("to_page", xmTextFieldWidgetClass,
						f2child,
						XmNcolumns, 5,
						XmNsensitive, False,
						XmNtopAttachment, XmATTACH_WIDGET,
						XmNtopWidget, range_marked_radio[act],
						XmNleftAttachment, XmATTACH_WIDGET,
						XmNleftWidget, to_label[act],
						NULL);
    XtAddCallback(page_to_text[act], XmNactivateCallback, cb_print, (XtPointer)act);

    str = XmStringCreateLocalized(ofstring);
    of_label[act] = XtVaCreateManagedWidget("of", xmLabelGadgetClass,
					    f2child,
					    XmNlabelString, str,
					    XmNsensitive, False,
					    XmNresizable, True,
					    XmNtopAttachment, XmATTACH_WIDGET,
					    XmNtopWidget, range_marked_radio[act],
					    XmNleftAttachment, XmATTACH_WIDGET,
					    XmNleftWidget, page_to_text[act],
					    NULL);
    XmStringFree(str);

    adjust_vertically(range_from_to_radio[act], page_from_text[act], DDIST);
    adjust_vertically(from_label[act], page_from_text[act], DDIST);
    adjust_vertically(to_label[act], page_from_text[act], DDIST);
    adjust_vertically(of_label[act], page_from_text[act], DDIST);

    XtManageChild(f2child);
    XtManageChild(frame2);

    /* box for the OK/Cancel button */
    box = XtVaCreateManagedWidget("box", xmFormWidgetClass,
				  pane,
				  XmNskipAdjust, True, /* don't resize this area */
				  NULL);
				  
    if (act == FILE_PRINT)
	str = XmStringCreateLocalized("Print");
    else
	str = XmStringCreateLocalized("Save");
    ok_button[act] = XtVaCreateManagedWidget("ok", xmPushButtonWidgetClass, box,
					     XmNlabelString, str,
					     XmNshowAsDefault, True,
					     XmNdefaultButtonShadowThickness, 1,
					     XmNtopAttachment, XmATTACH_FORM,
					     XmNbottomAttachment, XmATTACH_FORM,
					     XmNleftAttachment, XmATTACH_FORM,
					     /* to mimick appearance of native dialog buttons: */
					     XmNmarginWidth, 6,
					     XmNmarginHeight, 4,
					     XmNtopOffset, 10,
					     XmNbottomOffset, 10,
					     XmNleftOffset, 10,
					     NULL);
    XmStringFree(str);
    if (act == FILE_PRINT)
	XtAddCallback(ok_button[act], XmNactivateCallback, cb_print, (XtPointer)FILE_PRINT);
    else
	XtAddCallback(ok_button[act], XmNactivateCallback, cb_print, (XtPointer)FILE_SAVE);
    
    XtOverrideTranslations(ok_button[act], XtParseTranslationTable("<Key>Return:ArmAndActivate()"));

    str = XmStringCreateLocalized("Cancel");
    cancel_button[act] = XtVaCreateManagedWidget("cancel",
						 xmPushButtonWidgetClass, box,
						 XmNlabelString, str,
						 XmNdefaultButtonShadowThickness, 1,
						 XmNtopAttachment, XmATTACH_FORM,
						 XmNbottomAttachment, XmATTACH_FORM,
						 XmNrightAttachment, XmATTACH_FORM,
						 /* to mimick appearance of native dialog buttons: */
						 XmNmarginWidth, 6,
						 XmNmarginHeight, 4,
						 XmNtopOffset, 10,
						 XmNbottomOffset, 10,
						 XmNrightOffset, 10,
						 NULL);
    XmStringFree(str);

    if (act == FILE_PRINT)
	XtAddCallback(cancel_button[act], XmNactivateCallback, cb_print_cancel, NULL);
    else
	XtAddCallback(cancel_button[act], XmNactivateCallback, cb_save_cancel, NULL);

    { /* make the buttons the same size: */
	Dimension w1, w2;
	XtVaGetValues(ok_button[act], XtNwidth, &w1, NULL);
	XtVaGetValues(cancel_button[act], XtNwidth, &w2, NULL);
	if (w1 < w2)
	    XtVaSetValues(ok_button[act], XtNwidth, w2, NULL);
	else if (w2 > w1)
	    XtVaSetValues(cancel_button[act], XtNwidth, w1, NULL);
    }
    
    if (save_pr_info.range_radio < 0 && pageinfo_have_marked_pages()) {
	XmToggleButtonGadgetSetState(range_marked_radio[act], True, True);
    }

    XtManageChild(form);
    XtManageChild(pane);
    center_window(top_level_shell, top_level);
    XtMapWidget(top_level_shell);
    XmProcessTraversal(ok_button[act], XmTRAVERSE_CURRENT);
#endif /* not MOTIF */

    return top_level_shell;
}
#undef HORIZONTAL_RESIZING_NO
#undef HORIZONTAL_RESIZING_YES
#undef VERTICAL_RESIZING_NO
    

static void
set_target_radio(dialog_info_t *save_info, dialog_info_t *curr_info, print_or_save_action_t act)
{
#ifdef MOTIF
    if (save_info->print_radio > 0 && save_info->print_radio != curr_info->print_radio) {
	XmToggleButtonGadgetSetState(save_info->print_radio == PRINT_TO_PRINTER
				     ? to_printer_radio
				     : to_file_radio_or_label[act], True, True);
    }
#else /* MOTIF */
    UNUSED(act);
    UNUSED(curr_info);
    XawToggleSetCurrent(to_printer_radio, (XtPointer)save_info->print_radio);
#endif /* MOTIF */
}

static void
set_range_radio(dialog_info_t *save_info, dialog_info_t *curr_info, print_or_save_action_t act)
{
#ifdef MOTIF
    if (save_info->range_radio > 0 && save_info->range_radio != curr_info->range_radio) {
	Widget selected;
	switch (save_info->range_radio) {
	case PRINT_ALL:
	case SAVE_ALL:
	    selected = range_all_radio[act];
	    break;
	case PRINT_MARKED:
	case SAVE_MARKED:
	    selected = range_marked_radio[act];
	    break;
	case PRINT_RANGE:
	case SAVE_RANGE:
	    selected = range_from_to_radio[act];
	    break;
	default:
	    ASSERT(0, "wrong button type in set_range_radio()");
	}
	XmToggleButtonGadgetSetState(selected, True, True);
    }
#else /* MOTIF */
    UNUSED(curr_info);
    XawToggleSetCurrent(range_all_radio[act], (XtPointer)save_info->range_radio);
#endif /* MOTIF */
}

void
save_or_print_callback(print_or_save_action_t act)
{
#ifndef MOTIF
/*     static Dimension sw; /\* shell width *\/ */
#endif /* MOTIF */
    static char ofstring[128]; /* FIXME: allocate dynamically? */
    dialog_info_t *save_info = NULL, *curr_info = NULL;

    if (act == FILE_PRINT) {
	save_info = &save_pr_info;
	curr_info = &curr_pr_info;
	if (print_active) {
	    XRaiseWindow(DISP, XtWindow(print_shell));
	    return;
	}

	if (printlog_raise_active()) {
	    return;
	}
    }
    else if (act == FILE_SAVE) {
	save_info = &save_save_info;
	curr_info = &curr_save_info;
	if (save_active) {
	    XRaiseWindow(DISP, XtWindow(save_shell));
	    return;
	}
    }
    else {
	ASSERT(0, "wrong action type in save_or_print_callback()");
    }

    if (g_bak_fp == NULL) {
	do_popup_message(MSG_ERR, NULL, "No active dvi file");
	return;
    }

    if (pageno_correct == 1)
	sprintf(ofstring, "of %d", total_pages);
    else
	sprintf(ofstring, "of %d to %d", pageno_correct, total_pages + pageno_correct - 1);

    if (act == FILE_PRINT && print_shell == NULL)
	print_shell = create_print_or_save_window(act, ofstring);
    else if (act == FILE_SAVE && save_shell == NULL)
	save_shell = create_print_or_save_window(act, ofstring);
    else { /* if the window was already created */
#ifdef MOTIF
	XmString str;
#endif
	if (act == FILE_PRINT)
	    set_target_radio(save_info, curr_info, act);
	set_range_radio(save_info, curr_info, act);
#ifndef MOTIF
	XtVaSetValues(of_label[act], XtNlabel, ofstring, NULL);
	    
	if (act == FILE_PRINT) {
	    if (save_info->dvips_options != NULL) {
		XtVaSetValues(dvips_options_text, XtNstring, save_info->dvips_options, NULL);
		XawTextSetInsertionPoint(dvips_options_text, strlen(save_info->dvips_options));
	    }
	    else {
		XtVaSetValues(dvips_options_text, XtNstring, "", NULL);
	    }
	}
	    
	/* 	XtVaGetValues(print_shell, XtNwidth, &sw, NULL); */
#else /* not MOTIF */
	str = XmStringCreateLocalized(ofstring);
	XtVaSetValues(of_label[act], XmNlabelString, str, NULL);
	XmStringFree(str);

	if (act == FILE_PRINT) {
	    if (save_info->dvips_options != NULL) {
		XtVaSetValues(dvips_options_text, XmNvalue, save_info->dvips_options,
			      XmNcursorPosition, strlen(save_info->dvips_options), NULL);
	    }
	    else {
		XtVaSetValues(dvips_options_text, XmNvalue, "", NULL);
	    }
	}
#endif /* not MOTIF */
    }

    get_initial_dialog_values(save_info, act);

#ifndef MOTIF

    if (act == FILE_PRINT) {
	XtVaSetValues(print_to_printer_text, XtNstring, save_info->printer_name, NULL);
	XawTextSetInsertionPoint(print_to_printer_text, strlen(save_info->printer_name));
    }

    XtVaSetValues(print_to_file_text[act], XtNstring, save_info->file_name, NULL);
    XawTextSetInsertionPoint(print_to_file_text[act], strlen(save_info->file_name));

    if (act == FILE_PRINT) {
	XtVaSetValues(dvips_options_text, XtNstring, save_info->dvips_options, NULL);
	XawTextSetInsertionPoint(dvips_options_text, strlen(save_info->dvips_options));
    }
    
    XtVaSetValues(page_from_text[act], XtNstring, save_info->from_page, NULL);
    XawTextSetInsertionPoint(page_from_text[act], strlen(save_info->from_page));

    XtVaSetValues(page_to_text[act], XtNstring, save_info->to_page, NULL);
    XawTextSetInsertionPoint(page_to_text[act], strlen(save_info->to_page));
#else /* not MOTIF */
    if (act == FILE_PRINT)
	XtVaSetValues(print_to_printer_text, XmNvalue, save_info->printer_name,
		      XmNcursorPosition, strlen(save_info->printer_name), NULL);
    XtVaSetValues(print_to_file_text[act], XmNvalue, save_info->file_name,
		  XmNcursorPosition, strlen(save_info->file_name), NULL);
    if (act == FILE_PRINT)
	XtVaSetValues(dvips_options_text, XmNvalue, save_info->dvips_options,
		      XmNcursorPosition, strlen(save_info->dvips_options), NULL);
    XtVaSetValues(page_from_text[act], XmNvalue, save_info->from_page,
		  XmNcursorPosition, strlen(save_info->from_page), NULL);
    XtVaSetValues(page_to_text[act], XmNvalue, save_info->to_page,
		  XmNcursorPosition, strlen(save_info->to_page),  NULL);
#endif /* not MOTIF */
    
    if (act == FILE_PRINT) {
	XtPopup(print_shell, XtGrabNone);
	if (curr_pr_info.range_radio == PRINT_MARKED && !pageinfo_have_marked_pages()) {
#if MOTIF
	    XmToggleButtonGadgetSetState(range_all_radio[FILE_PRINT], True, True);
#endif
	    warn_no_marked_pages();
	}
	print_active = True;
    }
    else if (act == FILE_SAVE) {
	XtPopup(save_shell, XtGrabNone);
	if (curr_save_info.range_radio == SAVE_MARKED && !pageinfo_have_marked_pages()) {
#if MOTIF
	    XmToggleButtonGadgetSetState(range_all_radio[FILE_SAVE], True, True);
#endif
	    warn_no_marked_pages();
	}
	save_active = True;
    }
}

/*
 *	The actual printing
 */

static Widget print_confirm = NULL;


static void dvips_ended(int, struct xchild *);
static struct xchild print_child = {NULL, 0, True, "dvips", NULL, dvips_ended};

static char *read_from_dvips(int);
static void write_to_dvips(int);
static struct xio print_xio = { NULL, 0, XIO_IN,
#if HAVE_POLL
				NULL,
#endif
				read_from_dvips,
				write_to_dvips};

static void dvips_alarm(struct xtimer *);
static struct xtimer dvips_timer = {NULL, {0, 0}, XTM_DEFAULT, dvips_alarm
#if MOTIF_TIMERS
				    , NULL, NULL
#endif
};
static int dvips_status;
#define DVIPS_STAT_NONE 0
#define DVIPS_STAT_RUN 1
#define DVIPS_STAT_WAIT 2
static int dvips_sig;	/* SIGINT or SIGKILL */

static void print_do_it(print_or_save_action_t act);

static int
getpageno(Widget w)
{
    char *s, *p;
    int pageno;
#ifndef MOTIF
    XtVaGetValues(w, XtNstring, &s, NULL);
#else /* MOTIF */
    s = XmTextFieldGetString(w);
#endif
    p = s;
    if (*p == '-')
	++p;
    if (!isdigit((int)*p)) {
	return 0;
    }
    do {
	++p;
    } while (isdigit((int)*p));
    if (*p != '\0')
	return 0;

    pageno = atoi(s) - pageno_correct + 1;
#ifdef MOTIF
    XtFree(s);
#endif
    return pageno;
}


/*
  return text value of Widget w in newly allocated memory,
  free()ing old_val as appropriate. Returns NULL if
  text value starts with '\0'.
*/
static char *
get_saved_command(const char *old_val, Widget w)
{
    char *val = NULL;
    
    if (old_val != NULL) {
#if MOTIF
	XtFree((char *)old_val);
#else
	free((char*)old_val);
#endif
    }
#if MOTIF
    val = XmTextFieldGetString(w);
    if (*val == '\0') {
	XtFree((char *)val);
	val = NULL;
    }
#else
    XtVaGetValues(w, XtNstring, &val, NULL);
    if (*val == '\0') {
	val = NULL;
    }
    else {
	val = xstrdup(val);
    }
#endif
    return val;
}

/* fill curr_print_info and save_print_info with current values */
static void
get_save_window_state(dialog_info_t *save_info, dialog_info_t *curr_info, print_or_save_action_t act)
{
    save_info->print_radio = curr_info->print_radio;
    save_info->range_radio = curr_info->range_radio;

    if (act == FILE_PRINT) {
	save_info->printer_name = curr_info->printer_name
	    = get_saved_command(save_info->printer_name, print_to_printer_text);
    }
    save_info->file_name = curr_info->file_name = get_saved_command(save_info->file_name, print_to_file_text[act]);

    if (act == FILE_PRINT) {
	save_info->dvips_options = curr_info->dvips_options
	    = get_saved_command(save_info->dvips_options, dvips_options_text);
    }
    save_info->from_page = get_saved_command(save_info->from_page, page_from_text[act]);
    save_info->to_page = get_saved_command(save_info->to_page, page_to_text[act]);
    
    TRACE_GUI((stderr, "saved options:\nprinter: |%s|, file: |%s|, dvips: |%s|, from: |%s|, to: |%s|\n",
	       save_info->printer_name ? save_info->printer_name : "<NULL>", 
	       save_info->file_name ? save_info->file_name : "<NULL>",
	       save_info->dvips_options ? save_info->dvips_options : "<NULL>",
	       save_info->from_page ? save_info->from_page : "<NULL>",
	       save_info->to_page ? save_info->to_page : "<NULL>"));
}

static void
popdown_dialog_and_print(print_or_save_action_t act)
{
    if (act == FILE_PRINT)
	cb_print_cancel(NULL, NULL, NULL);
    else
	cb_save_cancel(NULL, NULL, NULL);
    print_do_it(act);
}

static void
print_precheck(print_or_save_action_t act)
{
    int page_range_begin, page_range_end;
    char *expanded_filename = NULL;
    
    /* Check for active confirm box */
    if (print_confirm != NULL) {
	XRaiseWindow(DISP, XtWindow(print_confirm));
	return;
    }

    /* Validate page range (if given) */
    if ((act == FILE_PRINT && curr_pr_info.range_radio == PRINT_RANGE)
	|| (act == FILE_SAVE && curr_save_info.range_radio == SAVE_RANGE)) {
	page_range_begin = getpageno(page_from_text[act]);
	page_range_end = getpageno(page_to_text[act]);
	if (page_range_begin <= 0) {
	    do_popup_message(MSG_ERR,
			     "Please specify a valid page range.",
			     "Invalid page range: start (%d) <= 0",
			     page_range_begin);
	    return;
	}
	else if (page_range_begin > page_range_end) {
	    do_popup_message(MSG_ERR,
			     "Please specify a valid page range.",
			     "Invalid page range: start (%d) > end (%d)",
			     page_range_begin, page_range_end);
	    return;
	}
	else if (page_range_end > total_pages) {
	    do_popup_message(MSG_ERR,
			     "Please specify a valid page range.",
			     "Invalid page range: end (%d) > total_pages (%d)",
			     page_range_end, total_pages);
	    return;
	}

	if (act == FILE_PRINT) {
	    curr_pr_info.from_page = xrealloc(curr_pr_info.from_page, length_of_int(page_range_begin) + 1);
	    curr_pr_info.to_page = xrealloc(curr_pr_info.to_page, length_of_int(page_range_end) + 1);
	    sprintf(curr_pr_info.from_page, "%d", page_range_begin);
	    sprintf(curr_pr_info.to_page, "%d", page_range_end);
	}
	else {
	    curr_save_info.from_page = xrealloc(curr_save_info.from_page, length_of_int(page_range_begin) + 1);
	    curr_save_info.to_page = xrealloc(curr_save_info.to_page, length_of_int(page_range_end) + 1);
	    sprintf(curr_save_info.from_page, "%d", page_range_begin);
	    sprintf(curr_save_info.to_page, "%d", page_range_end);
	}
    }

    if (act == FILE_PRINT)
	get_save_window_state(&save_pr_info, &curr_pr_info, act);
    else
	get_save_window_state(&save_save_info, &curr_save_info, act);

    { /* expand and canonicalize path name */
	const char *fname = act == FILE_PRINT ? save_pr_info.file_name : save_save_info.file_name;
	expanded_filename = expand_homedir(fname);

	if (expanded_filename == NULL) {
	    do_popup_message(MSG_WARN,
			     "Please specify either a filename or a full path, without using \"~\" or \"~user\".",
			     "Couldn't expand filename \"%s\" to a full path.",
			     fname);
	    return;
	}

	if (act == FILE_PRINT) {
	    free(save_pr_info.expanded_filename);
	    save_pr_info.expanded_filename = xstrdup(expanded_filename);
	}
	else {
	    free(save_save_info.expanded_filename);
	    save_save_info.expanded_filename = xstrdup(expanded_filename);
	}
    }


    if (act == FILE_SAVE) {
	char *tmp = expand_filename_append_dvi(save_save_info.expanded_filename, USE_DVI_PATH);
	if (strcmp(tmp, dvi_name) == 0) {
	    do_popup_message(MSG_ERR,
			     NULL,
			     "You cannot overwrite the current DVI file (%s). "
			     "Please select a different file name.",
			     save_save_info.expanded_filename);
	    free(tmp);
	    return;
	}
    }
    if (act == FILE_SAVE || (act == FILE_PRINT && save_pr_info.print_radio == PRINT_TO_FILE)) {
	/* check whether to clobber existing file */
	struct stat statbuf;
	if (stat(expanded_filename, &statbuf) == 0 && S_ISREG(statbuf.st_mode)) {
	    do_popup_ok_or_cancel_dialog(NULL,
					 "Replace",
					 /* if user selects OK */
					 popdown_dialog_and_print, (XtPointer)act,
					 /*
					   do nothing if user selects `cancel'; this will return to the
					   printing dialog, since user hasn't changed their mind about
					   printing, but would probably just like to choose another filename
					 */
					 NULL, NULL,
					 "%s already exists.\nDo you want to replace it?",
					 expanded_filename);
	    return;
	}
    }
    
    popdown_dialog_and_print(act);
}

static void
cb_dvips_unkeep(Widget w, XtPointer client_data, XtPointer call_data)
{

    UNUSED(w);
    UNUSED(client_data);
    UNUSED(call_data);
    
    /* use negative value so that we don't loose the value itself */
    TRACE_GUI((stderr, "toggling resource from %d to %d, %d to %d",
	       resource.dvips_hang, -resource.dvips_hang,
	       resource.dvips_fail_hang, -resource.dvips_fail_hang));
    resource.dvips_hang = -resource.dvips_hang;
    resource.dvips_fail_hang = -resource.dvips_fail_hang;
    if (dvips_status == DVIPS_STAT_WAIT) {
	dvips_status = DVIPS_STAT_NONE;
	cancel_timer(&dvips_timer);
    }
}

/* if user selects the window manager destroy button */
static void
cb_dvips_destroy(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(client_data);
    UNUSED(call_data);

    if (dvips_status == DVIPS_STAT_RUN) {
	kill(print_child.pid, dvips_sig);
	dvips_sig = SIGKILL;
	printlog_append("^C", 2);
    }

    if (dvips_status == DVIPS_STAT_WAIT) {
	dvips_status = DVIPS_STAT_NONE;
	cancel_timer(&dvips_timer);
    }

    printlog_reset();
    printlog_popdown(True);
}
    
static void
cb_dvips_cancel(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(client_data);
    UNUSED(call_data);

    if (dvips_status != DVIPS_STAT_RUN)
	return;	/* How did we get here? */

    kill(print_child.pid, dvips_sig);
    dvips_sig = SIGKILL;
    printlog_append("^C", 2);
}

static void
cb_dvips_close(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(client_data);
    UNUSED(call_data);

    if (dvips_status == DVIPS_STAT_RUN)
	return;	/* How did we get here? */

    if (dvips_status == DVIPS_STAT_WAIT) {
	dvips_status = DVIPS_STAT_NONE;
	cancel_timer(&dvips_timer);
    }

    printlog_popdown(True);
}

/* FIXME: implement this in terms of fork_process:
   fork_process(argv[0], argv);
   Currently it doesn't work since the DVI file and the
   output file are realized via redirecting stdin/stdout.
   Is this neccessary?
   It surely obscures the log output (a lot of users won't
   understand the I/O redirection).
   
   FIXME beta: Remove the detailed output, and implement
   a progress window that only shows the page numbers.
*/

/*
  f != NULL: not printing to printer, FILE * to file being created
  tmp_file != NULL: printing selected pages, temporary DVI file containing
		    the selected pages
 */
static void
fork_dvips(char **argv, FILE *f, FILE *tmp_file)
{
    int print_io[2];
    int i;
    
/*     printlog_append(argv[0], strlen(argv[0])); */

    printlog_popup();

    printlog_append("Calling: ", strlen("Calling: "));
    printlog_append(argv[0], strlen(argv[0]));
    printlog_append(" with arguments:\n", strlen(" with arguments:\n"));

    for (i = 1; argv[i] != NULL; i++) {
	char argstr[64];
	sprintf(argstr, "%d: |", i);
	printlog_append(argstr, strlen(argstr));
	printlog_append(argv[i], strlen(argv[i]));
	printlog_append("|\n", 2);
    }
    
    if (xpipe(print_io) != 0) {
 	perror("[xdvi] pipe");
 	return;
    }
 
    /* Fork process */
    
    /* flush output buffers to avoid double buffering (i.e. data
       waiting in the output buffer being written twice, by the parent
       and the child) */
    fflush(stderr);
    fflush(stdout);
    
    print_child.name = xstrdup("dvips");
    print_child.pid = vfork();
    if (print_child.pid == 0) {		/* if child */
	if (tmp_file != NULL) { /* if printing from temporary instead of main file */
	    (void)dup2(fileno(tmp_file), 0);
	    (void)close(fileno(tmp_file));
	}
	else {
	    int fd;
	    /* open dvi file anew so that seek()s on the
	       original FILE * in the parent don't affect us */
	    (void)close(0);
	    (void)close(fileno(g_bak_fp));
	    if ((fd = open(dvi_name, O_RDONLY)) < 0 && (errno == EMFILE || errno == ENFILE)) {
		close_a_file();
		fd = open(dvi_name, O_RDONLY);
	    }
	    if (fd < 0) {
		perror(dvi_name);
		fflush(stderr);
		_exit(1);
	    }
	    if (fd > 0) {
		dup2(fd, 0);
		(void)close(fd);
	    }
	}
	(void)lseek(0, 0, SEEK_SET);
 	(void)close(1);
 	if (f == NULL) { /* printing to printer */
 	    (void)dup2(print_io[1], 1);
 	}
 	else {	/* printing to file */
 	    (void)dup2(fileno(f), 1);
 	    (void)close(fileno(f));
 	}
 	(void)close(2);
 	(void)dup2(print_io[1], 2);
 	(void)close(print_io[1]);
 	(void)close(print_io[0]);
 
 	if (setsid() == -1) {	/* so we can kill the process group */
 	    perror("setsid");
 	    fflush(stderr);
 	    _exit(1);
 	}
 	(void)execvp(*argv, argv);
 	do_popup_message(MSG_ERR,
			 NULL,
			 "Execution of \"%s\" failed: %s.\n", *argv, strerror(errno));
 	fflush(stderr);
 	_exit(1);
    }
 
    if (f != NULL)
 	fclose(f);
 
    if (print_child.pid == -1) {	/* error */
 	perror("[xdvi] vfork");
 	return;
    }
 
    set_chld(&print_child);
    dvips_sig = SIGINT;

    (void)close(print_io[1]);
 
    /* Set up file descriptor for non-blocking I/O */
    prep_fd(print_io[0], True);
    print_xio.fd = print_io[0];
    set_io(&print_xio);
 
    dvips_status = DVIPS_STAT_RUN;	/* running */
}

static char **
create_dvips_argv(dialog_info_t *info, FILE **tmp_dvi_file)
{
    size_t argv_len = 32; /* should be ample ... */
    char **argv = xmalloc(argv_len * sizeof *argv);
    size_t idx = 0;
    argv[idx++] = xstrdup(resource.dvips_path);

    if (info->dvips_options != NULL) {
	char **dvips_args = get_space_separated_list(info->dvips_options);
	int i;
	for (i = 0; dvips_args[i] != NULL;) {
	    argv[idx++] = dvips_args[i++];
	}
    }
    
    argv[idx++] = xstrdup("-f");
    
    if (info->print_radio == PRINT_TO_PRINTER) {
	char **printer_args = get_space_separated_list(info->printer_name);
	int i;
	ASSERT(*printer_args != NULL, "args should contain at least the string \"lpr\"");
	argv[idx] = xstrdup("-o!");
	argv[idx] = xstrcat(argv[idx], printer_args[0]);
	/* now append everything to the printer pipe command, so that dvips doesn't see it as an option */
	for (i = 1; printer_args[i] != NULL; i++) {
	    argv[idx] = xstrcat(argv[idx], " ");
	    argv[idx] = xstrcat(argv[idx], printer_args[i]);
	}
	idx++;
    }

    if (info->range_radio == PRINT_MARKED) {
	char *tmpfile = NULL;
	int tmp_fd = xdvi_temp_fd(&tmpfile);
/* 	fprintf(stderr, "temporary file name: |%s|, %d\n", tmpfile, tmp_fd); */
	if (tmp_fd == -1) {
	    do_popup_message(MSG_ERR,
			     "Your kernel thinks it has run out of "
			     "file descriptors. Close a few applications, then try again.",
			     "Couldn't create temporary DVI file for printing: %s", strerror(errno));
	    return NULL;
	}
	/* don't use XFOPEN here, since we don't want to treat an error in opening the file as fatal. */
	/* `b' in mode won't hurt ... */
	if ((*tmp_dvi_file = fdopen(tmp_fd, "wb+")) == NULL && (errno == EMFILE || errno == ENFILE)) {
	    close_a_file();
	    *tmp_dvi_file = fdopen(tmp_fd, "wb+");
	}
	if (*tmp_dvi_file == NULL) {
	    do_popup_message(MSG_ERR,
			     "Xdvi needs to create a temporary file containing the "
			     "currently marked pages, but creating that file failed. "
			     "Try to fix the cause of the problem, or choose "
			     "\"Print selected pages\" as a workaround.",
			     "Couldn't open temp file \"%s\" for writing: %s\n",
			     tmpfile, strerror(errno));
	    return NULL;
	}
	else {
	    /* print selected pages to temp file */
	    if (!select_pages(*tmp_dvi_file, check_marked, NULL, NULL)) {
		do_popup_message(MSG_ERR, NULL, "Couldn't create DVI file with marked pages");
		return NULL;
	    }

	    /* position at beginning of file */
	    fseek(*tmp_dvi_file, 0, SEEK_SET);
	    if (debug & DBG_GUI) {
		/* preserve it, for debugging */
		fprintf(stderr, "created temp DVI file %s (NOT removing it afterwards)\n", tmpfile);
	    }
	    else {
		/* so that it gets removed automatically after it has been closed */
		unlink(tmpfile);
	    }
	}
    }
    else if (info->range_radio == PRINT_RANGE) {
	ASSERT(info->from_page != NULL && info->to_page != NULL, "Range not initialized");
	argv[idx] = xstrdup("-p=");
	argv[idx] = xstrcat(argv[idx], info->from_page);
	idx++;
	argv[idx] = xstrdup("-l=");
	argv[idx] = xstrcat(argv[idx], info->to_page);
	idx++;
    }
    argv[idx] = NULL; /* terminate it */
    ASSERT(idx <= argv_len, "Too many elements");
    return argv;
}

static void
print_do_it(print_or_save_action_t act)
{
    FILE *fp = NULL;
    
    if (g_bak_fp == NULL) {
	do_popup_message(MSG_ERR, NULL, "No active dvi file");
	return;
    }

    /* Open file for writing (if necessary) */
    if ((act == FILE_PRINT && save_pr_info.print_radio == PRINT_TO_FILE)
	|| act == FILE_SAVE) {
	char *fname = act == FILE_PRINT ? save_pr_info.expanded_filename : save_save_info.expanded_filename;
	ASSERT(fname != NULL, "Filename not initialized properly");
	
	if ((fp = XFOPEN(fname, "w")) == NULL) {
	    do_popup_message(MSG_ERR,
			     NULL, "Could not open %s for writing: %s.", fname,
			     strerror(errno));
	    return;
	}
    }

    if (act == FILE_PRINT) { /* printing DVI file or parts of it */
	FILE *tmp_dvi_file = NULL; /* temporary DVI file containing selected pages */
	char **argv;
	printlog_create(cb_dvips_close, cb_dvips_cancel, cb_dvips_destroy, cb_dvips_unkeep);
    
	argv = create_dvips_argv(&save_pr_info, &tmp_dvi_file); /* this eventually creates tmp_dvi_file */
	if (argv == NULL) {
	    /* something went *really* wrong; assume user has already been warned about it */
	    return;
	}
    
	/* sanity check. Can we de-mangle this further? */
	ASSERT((fp == NULL && save_pr_info.print_radio == PRINT_TO_PRINTER)
	       || (fp != NULL && save_pr_info.print_radio == PRINT_TO_FILE), "");
	
	fork_dvips(argv, fp, tmp_dvi_file);
	
	/* dellocate argv */
	{
	    int i = 0;
	    for (i = 0; argv[i] != NULL; i++) {
		free(argv[i]);
	    }
	    free(argv);
	}
    }
    else { /* saving DVI file or parts of it*/
	Boolean retval = False;
	struct specials_stack_t warn_included_files = { 0, NULL };

	ASSERT(fp != NULL, "File pointer not initialized properly");
	
	if (save_save_info.range_radio == SAVE_MARKED) {
	    retval = select_pages(fp, check_marked, NULL, &warn_included_files);
	    fclose(fp);
	}
	else if (save_save_info.range_radio == SAVE_RANGE) {
	    struct select_pages_t sel_info;
	    /* convert to zero-based numbers */
	    sel_info.from = strtoul(save_save_info.from_page, (char **)NULL, 10) - 1;
	    sel_info.to = strtoul(save_save_info.to_page, (char **)NULL, 10) - 1;
	    retval = select_pages(fp, check_pagerange, &sel_info, &warn_included_files);
	    /*  	    retval = select_pagerange(fp, from, to, &warn_included_files); */
	    fclose(fp);
	}
	else {
	    retval = copy_file(fp, &warn_included_files);
	    fclose(fp);
	}
	if (retval) {
	    if (warn_included_files.stack_len > 0) {
		char *warn_files = xstrdup("");
		size_t i;
		for (i = 0; i < warn_included_files.stack_len; i++) {
		    warn_files = xstrcat(warn_files, warn_included_files.items[i].content);
		    warn_files = xstrcat(warn_files, "\n");
		}
		do_popup_message(MSG_INFO,
				 "The files listed in the message (probably images) are not part "
				 "of the newly created DVI file; they are only referenced by that file. "
				 "Therefore, if you want to distribute this file or use it on a different computer,"
				 "you will need to distribute all the referenced files, too.\n\n"
				 "Tip: A better format for distributing files is Postscript or PDF. "
				 "To create a Postscript file, use the `File->Print' menu, "
				 "then select `Print to file' in the dialog window.",
				 "Created %s.\n"
				 "Please note that the following files are referenced by this file, "
				 "and are needed for displaying or printing it correctly:\n%s\n",
				 save_save_info.expanded_filename, warn_files);
		free(warn_files);
	    }
	    else {
		print_statusline(STATUS_LONG, "Pages written to %s.", save_save_info.file_name);
	    }
	}
	else {
	    do_popup_message(MSG_INFO,
			     NULL,
			     "There was an error writing to %s: %s.", save_save_info.file_name, strerror(errno));
	}
    }
}



static char *
read_from_dvips(int ignored)
{
    int bytes;
    char line[80];

    UNUSED(ignored);
    
    for (;;) {
#ifndef MOTIF
	bytes = read(print_xio.fd, line, sizeof line);
#else
	bytes = read(print_xio.fd, line, sizeof line - 1);
#endif
	if (bytes < 0) {
	    if (AGAIN_CONDITION)
		break;
	    perror("xdvi: read_from_dvips");
	    break;
	}

	if (bytes == 0)
	    break;
	else {
#ifdef MOTIF
	    line[bytes] = '\0';
#endif
	    printlog_append(line, bytes);
	}
    }
    return NULL; /* FIXME */
}

static void
write_to_dvips(int ignored)
{
    UNUSED(ignored);
    
    return;
}



static void
dvips_ended(int status, struct xchild *should_be_child)
{
    char str[128] = "";
    int ms;

    UNUSED(should_be_child);
    
    read_from_dvips(0);
    clear_io(&print_xio);
    (void)close(print_xio.fd);

    if (WIFEXITED(status)) {
	if (WEXITSTATUS(status) == 0) {
	    if (save_pr_info.print_radio == PRINT_TO_FILE) {
		printlog_append("Created Postscript file ", strlen("Created Postscript file "));
		printlog_append(save_pr_info.expanded_filename, strlen(save_pr_info.expanded_filename));
		printlog_append(".\n", 2);
	    }
	    else {
		printlog_append("Done.\n", strlen("Done.\n"));
	    }
	}
	else
	    sprintf(str, "\nPrint process returned exit code %d.\n",
		    WEXITSTATUS(status));
    }
    else if (WIFSIGNALED(status))
	sprintf(str, "\nPrint process terminated by signal %d.\n",
		WTERMSIG(status));
    else
	sprintf(str, "\nPrint process returned unknown status 0x%x.\n",
		status);

    ms = resource.dvips_hang;
    if (str[0] != '\0') {
	XBell(DISP, 0);
	ms = resource.dvips_fail_hang;
	printlog_append(str, strlen(str));
    }

    if (ms > 0) {
	set_timer(&dvips_timer, ms);
	dvips_status = DVIPS_STAT_WAIT;
    }
    else {
	dvips_status = DVIPS_STAT_NONE;
    }

    printlog_enable_closebutton();
}


static void
dvips_alarm(struct xtimer *should_be_timer)
{
    UNUSED(should_be_timer);

    printlog_popdown(False);
    
    dvips_status = DVIPS_STAT_NONE;
}

