/*
 * Copyright (c) 2001 Marcin Dalecki
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * PAUL VOJTA OR ANY OTHER AUTHOR OF THIS SOFTWARE BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 */

#include "xdvi-config.h"
#include "xdvi.h"

#include "xm_menu.h"
#include "xm_toolbar.h"
#include "util.h"
#include "statusline.h"
#include "c-openmx.h"

#ifdef MOTIF /* needed for `make depend' */

#include <Xm/RowColumn.h>
#include <Xm/Frame.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/CascadeB.h>
#include <Xm/ToggleB.h>
#include <Xm/ToggleBG.h>
#include <Xm/Separator.h>

#define MAX_MENU_CNT 128
static Widget menus[MAX_MENU_CNT];

/*
 * Menu bar implementation for the Motif widget set.
 */


/* structures for menu data */
struct pulldown_item_info {
    char *title;
    XmButtonType type;		/* button type - one of XmPUSHBUTTON, XmSEPARATOR, XmRADIOBUTTON, XmCHECKBUTTON
				   (see XmNbuttonType in XmRowColumn(3)) */
    char mnemonic;		/* 0 if none */
    char *accelerator;		/* NULL if none */
    char *proc;			/* name of action procedure to call */
    char *args;			/* arguments for callback procedure */
};


struct menu_item_info {
    char *title;
    char mnemonic;
    XtCallbackProc callback;
    struct pulldown_item_info *pulldown_item_info;
};

struct callback_info {
    char *function;
    char *args;
};

/* FIXME:
   The menu accelerators conflict with the key translations, e.g. when using
   `n' for `Next page', xdvi will skip forward 2 pages.
 */
static struct pulldown_item_info file_pulldown_info[] = {
    { "Open",		XmPUSHBUTTON,	'O',	"Ctrl+F",	"select-dvi-file",	NULL },
    { "Reload",		XmPUSHBUTTON,	'R',	"R",		"reread-dvi-file",	NULL },
    { "Print ...",	XmPUSHBUTTON,	'P',	NULL,		"print",		NULL },
    { "Save ...",	XmPUSHBUTTON,	'S',	NULL,		"save",			NULL },
    { "Sep",		XmSEPARATOR,	 0,	NULL,		NULL,			NULL },
    { "Exit",		XmPUSHBUTTON,	'x',	"q",		"quit",			NULL },
    { NULL,		0,		 0,	NULL,		NULL,			NULL }
};

static struct pulldown_item_info navigate_pulldown_info[] = {
    { "Page - 10",	XmPUSHBUTTON, 0,	"10 p",		"back-page",		"10" },
    { "Page - 5",	XmPUSHBUTTON, 0,	"5 p",		"back-page",		"5"  },
    { "Previous Page",	XmPUSHBUTTON, 'P',	"Backspace",	"back-page",		"1"  },
    { "Next Page",	XmPUSHBUTTON, 'N',	"Return",	"forward-page",		"1"  },
    { "Page + 5",	XmPUSHBUTTON, 0,	"5 n",		"forward-page",		"5"  },
    { "Page + 10",	XmPUSHBUTTON, 0,	"10 n",		"forward-page",		"10" },
    { "Sep",		XmSEPARATOR,  0,	NULL,		NULL,			NULL },
    { "First Page",	XmPUSHBUTTON, 'F',	"1 g",		"goto-page",		"1"  },
    { "Last Page",	XmPUSHBUTTON, 'L',	"g",		"goto-page",		"e"  },
    { "Sep",		XmSEPARATOR,  0,	NULL,		NULL,			NULL },
    { "Back Hyperlink",	XmPUSHBUTTON, 'B',	"B",		"htex-back",		NULL },
    { "Forward Hyperlink",	XmPUSHBUTTON, 'F',	"F",	"htex-forward",		NULL },
    { NULL,		0,	       0,	NULL,		NULL,			NULL }
};

static struct pulldown_item_info mark_pulldown_info[] = {
    { "Mark All Pages",	XmPUSHBUTTON, 'M',	NULL,		"toggle-mark",		"-1" },
    { "Unmark All Pages",	XmPUSHBUTTON, 'U',	"0 m",		"toggle-mark",		"0"  },
    { "Sep",		XmSEPARATOR,  'U',	NULL,		NULL,			NULL },
    { "Toggle Odd Pages",	XmPUSHBUTTON, 'O',	"1 m",		"toggle-mark",		"1"  },
    { "Toggle Even Pages",	XmPUSHBUTTON, 'E',	"2 m",		"toggle-mark",		"2"  },
    { "Sep",		XmSEPARATOR,  0,	NULL,		NULL,			NULL },
    { "Toggle Current Page",	XmPUSHBUTTON, 'C',	"m",		"toggle-mark",		NULL },
    { NULL,		0,	       0,	NULL,		NULL,			NULL }
};

static struct pulldown_item_info scale_pulldown_info[] = {
    { "Shrink by 1",	XmRADIOBUTTON, '1',	"1 s",	"set-shrink-factor",	"1"  },
    { "Shrink by 2",	XmRADIOBUTTON, '2',	"2 s",	"set-shrink-factor",	"2"  },
    { "Shrink by 3",	XmRADIOBUTTON, '3',	"3 s",	"set-shrink-factor",	"3"  },
    { "Shrink by 4",	XmRADIOBUTTON, '4',	"4 s",	"set-shrink-factor",	"4"  },
    { "Shrink by 5",	XmRADIOBUTTON, '5',	"5 s",	"set-shrink-factor",	"5"  },
    { "Shrink by 6",	XmRADIOBUTTON, '6',	"5 s",	"set-shrink-factor",	"6"  },
    { "Shrink by 7",	XmRADIOBUTTON, '7',	"7 s",	"set-shrink-factor",	"7"  },
    { "Shrink by 8",	XmRADIOBUTTON, '8',	"8 s",	"set-shrink-factor",	"8"  },
    { "Shrink by 9",	XmRADIOBUTTON, '9',	"9 s",	"set-shrink-factor",	"9"  },
    { "Sep",		XmSEPARATOR,   0,	NULL,	NULL,			NULL },
    { "Fit in Window",	XmPUSHBUTTON,  'F',	"s",	"set-shrink-factor",	NULL },
    { "Sep",		XmSEPARATOR,  0,	NULL,		NULL,			NULL },
    { "Zoom In",	XmPUSHBUTTON, 'I',	"Ctrl-+",		"set-shrink-factor", "+" },
    { "Zoom Out",	XmPUSHBUTTON, 'O',	"Ctrl--",		"set-shrink-factor", "-" },
    { NULL,		0,		 0,	NULL,	NULL,			NULL }
};

static struct pulldown_item_info options_pulldown_info[] = {
    { "Show PS",	XmRADIOBUTTON,	 'S',	"1v",	"set-ps",		"1" },
    { "Show PS and Bounding Boxes",	XmRADIOBUTTON,	 'B',	"2v",	"set-ps",		"2" },
    { "Show Bounding Boxes Only",	XmRADIOBUTTON,	 'O',	"0v",	"set-ps",		"0" },
    { "Sep",		XmSEPARATOR,   0,	NULL,	NULL,			NULL },
    { "Use PS Anti-Aliasing",	XmCHECKBUTTON,	 'A',	"V",	"set-gs-alpha",		"toggle" },
    { "Sep",		XmSEPARATOR,   0,	NULL,	NULL,			NULL },
    { "Ruler Mode",	XmCHECKBUTTON,	 'R',	"Ctrl-m",	"ruler-mode",		"toggle" },
    { "Keep Position",	XmCHECKBUTTON,	 'K',	"k",	"set-keep-flag",	"toggle" },
    { "Use TeX Page Numbers",	XmCHECKBUTTON, 'T', "T",	"use-tex-pages","toggle" },
    { "Sep",		XmSEPARATOR,   0,	NULL,	NULL,			NULL },
    { "Show Statusline",XmCHECKBUTTON,   'S',	"1x", "set-expert-mode",	"1" },
    { "Show Scrollbars",XmCHECKBUTTON,   'C',	"2x", "set-expert-mode",	"2" },
    { "Show Pagelist",	XmCHECKBUTTON,   'P',	"3x", "set-expert-mode",	"3" },
    { "Show Toolbar",	XmCHECKBUTTON,   'T',	"4x", "set-expert-mode",	"4" },
    /* Note: we don't expose `show menubar', since users might not know how to get it back! */
    { "Sep",		XmSEPARATOR,   0,	NULL,	NULL,			NULL },
    { "Fonts Darker",	XmPUSHBUTTON,   'D',	"Alt-Ctrl-+", "change-density",	"25" },
    { "Fonts Lighter",	XmPUSHBUTTON,   'L',	"Alt-Ctrl--", "change-density",	"-25" },
    { NULL,		0,		 0,	NULL,	NULL,			NULL }
};
    
static struct pulldown_item_info help_pulldown_info[] = {
    { "Introduction...",	XmPUSHBUTTON,	'I',	NULL,	"help",	"Introduction"	},
    { "Page Motion ...",	XmPUSHBUTTON,	'P',	NULL,	"help",	"Page Motion"	},
    { "Other Commands ...",	XmPUSHBUTTON,	'O',	NULL,	"help",	"Other Commands"	},
    { "Hyperlinks ...",		XmPUSHBUTTON,	'H',	NULL,	"help",	"Hyperlinks"		},
    { "Mouse Buttons ...",	XmPUSHBUTTON,	'M',	NULL,	"help",	"Mouse Buttons"	},
    { "Printing and Saving ...",	XmPUSHBUTTON,	'M',	NULL,	"help",	"Printing and Saving"	},
    { "Source Specials ...",	XmPUSHBUTTON,	'S',	NULL,	"help",	"Source Specials"	},
    { "Sep",		XmSEPARATOR,   0,	NULL,	NULL,			NULL },
    { "New features in this version",	XmPUSHBUTTON,	'F',	NULL,	"load-url",
	"http://sourceforge.net/project/shownotes.php?release_id=168307"	},
    { "Send a Bug Report",	XmPUSHBUTTON,	'B',	NULL,	"load-url",
	"http://sourceforge.net/tracker/?func=add&group_id=23164&atid=377580"	},
    { NULL,		0,		 0,	NULL,	NULL,			NULL }
};

#define MAX_CNT 20 /* maximum number of menu entries */

/* create menus from data */
static void
create_pulldown_menu(int pos, Widget parent, struct menu_item_info menu_info)
{
    struct pulldown_item_info *item;
    size_t i;
    int n;
    size_t alloc_step = 20;
    size_t alloc_len = 0;
    Arg *args = NULL;

    XmButtonTypeTable pulldownMenuButtonTypes = NULL;
    XmStringTable pulldownMenuStrings = NULL;
    XmKeySymTable pulldownMenuMnemonics = NULL;
    String *pulldownMenuAccelerators = NULL;
    XmStringTable pulldownMenuAcceleratorText = NULL;

    ASSERT(pos < MAX_MENU_CNT, "index out of range");
    
    item = menu_info.pulldown_item_info;
    
    for (i = 0; item[i].title != NULL; i++) {
	/* (re)allocate structures if neccessary */
	if (i >= alloc_len) {
	    alloc_len += alloc_step;
	}
	/* I wonder where that sick idea of having the Xt*alloc functions return char * comes from ... */
	pulldownMenuButtonTypes = (void *)XtRealloc((void *)pulldownMenuButtonTypes, sizeof (XmButtonType *) * alloc_len);
	pulldownMenuStrings = (void *)XtRealloc((void *)pulldownMenuStrings, sizeof (XmString *) * alloc_len);
	pulldownMenuMnemonics = (void *)XtRealloc((void *)pulldownMenuMnemonics, sizeof (KeySym *) * alloc_len);
	pulldownMenuAccelerators = (void *)XtRealloc((void *)pulldownMenuAccelerators, sizeof (String *) * alloc_len);
	pulldownMenuAcceleratorText = (void *)XtRealloc((void *)pulldownMenuAcceleratorText, sizeof (XmString *) * alloc_len);
	args = (void *)XtRealloc((void *)args, sizeof *args * alloc_len);
	
	pulldownMenuButtonTypes[i] = item[i].type;
	pulldownMenuMnemonics[i] = item[i].mnemonic;
	pulldownMenuAccelerators[i] = NULL;
	/* FIXME: There's something weird with the memory allocation scheme here;
	   identical strings in item[i].title will case a segfault!
	   Prevent by checking explicitly for "Sep", for the time being.
	*/
	if (item[i].title != NULL && strcmp(item[i].title, "Sep") != 0)
	    pulldownMenuStrings[i] = XmCvtCTToXmString(item[i].title);
	else
	    pulldownMenuStrings[i] = NULL;
	if (item[i].accelerator != NULL)
	    pulldownMenuAcceleratorText[i] = XmCvtCTToXmString(item[i].accelerator);
	else
	    pulldownMenuAcceleratorText[i] = NULL;
    }
    n = 0;
    XtSetArg(args[n], XmNsimpleCallback, menu_info.callback);			n++;
    XtSetArg(args[n], XmNpostFromButton, pos);					n++;
    XtSetArg(args[n], XmNbuttonCount, i);					n++;
    XtSetArg(args[n], XmNbuttonType, pulldownMenuButtonTypes);			n++;
    XtSetArg(args[n], XmNbuttons, pulldownMenuStrings);				n++;
    XtSetArg(args[n], XmNbuttonMnemonics, pulldownMenuMnemonics);		n++;
    XtSetArg(args[n], XmNbuttonAccelerators, pulldownMenuAccelerators);		n++;
    XtSetArg(args[n], XmNbuttonAcceleratorText, pulldownMenuAcceleratorText);	n++;
    
    menus[pos] = XmCreateSimplePulldownMenu(parent,
					    item[i].title == NULL ? "sep" : item[i].title,
					    args, n);

    /* free memory */
    for (i = 0; item[i].title != NULL; i++) {
	if (pulldownMenuStrings[i] != NULL)
	    XmStringFree(pulldownMenuStrings[i]);
	if (pulldownMenuAcceleratorText[i] != NULL)
	    XmStringFree(pulldownMenuAcceleratorText[i]);
    }
    XtFree((void *)pulldownMenuButtonTypes);
    XtFree((void *)pulldownMenuStrings);
    XtFree((void *)pulldownMenuMnemonics);
    XtFree((void *)pulldownMenuAccelerators);
    XtFree((void *)pulldownMenuAcceleratorText);
}

/* since the struct contains also separators, but these are not counted in the
   index passed in the callback, we need to compute the real position in the struct
   (counting also the separators)
*/
static size_t
get_index(struct pulldown_item_info *info, int idx)
{
    int i, offset = 0;
    for (i = 0; info[offset].title != 0 && i < idx; i++, offset++) {
	if (info[offset].type == XmSEPARATOR)
	    offset++;
    }
    /* account for trailing separators */
    while (info[offset].type == XmSEPARATOR) {
	offset++;
    }

    return offset;
}

static void
invoke_callback(Widget w, XmAnyCallbackStruct *call_data, struct pulldown_item_info item)
{
    char *params[1];
    if (item.proc != NULL) {
	if (debug & DBG_GUI)
	    fprintf(stderr, "invoking proc: |%s| with args: |%s|\n", item.proc, item.args);
	if (item.args != NULL) {
	    params[0] = item.args;
	    XtCallActionProc(w, item.proc, call_data->event, params, 1);
	}
	else
	    XtCallActionProc(w, item.proc, call_data->event, NULL, 0);
    }
}

static void
file_pulldown_callback(Widget w, XtPointer client_data, XmAnyCallbackStruct *call_data)
{
    int i = get_index(file_pulldown_info, (int)client_data);
    invoke_callback(w, call_data, file_pulldown_info[i]);
}

static void
navigate_pulldown_callback(Widget w, XtPointer client_data, XmAnyCallbackStruct *call_data)
{
    int i = get_index(navigate_pulldown_info, (int)client_data);
    invoke_callback(w, call_data, navigate_pulldown_info[i]);
}

static void
scale_pulldown_callback(Widget w, XtPointer client_data, XmAnyCallbackStruct *call_data)
{
    int i = get_index(scale_pulldown_info, (int)client_data);
    invoke_callback(w, call_data, scale_pulldown_info[i]);
}

static void
mark_pulldown_callback(Widget w, XtPointer client_data, XmAnyCallbackStruct *call_data)
{
    int i = get_index(mark_pulldown_info, (int)client_data);
    invoke_callback(w, call_data, mark_pulldown_info[i]);
}

static void
options_pulldown_callback(Widget w, XtPointer client_data, XmAnyCallbackStruct *call_data)
{
    int i = get_index(options_pulldown_info, (int)client_data);
    invoke_callback(w, call_data, options_pulldown_info[i]);
}

static void
help_pulldown_callback(Widget w, XtPointer client_data, XmAnyCallbackStruct *call_data)
{
    int i = get_index(help_pulldown_info, (int)client_data);
    invoke_callback(w, call_data, help_pulldown_info[i]);
}


/* FIXME: remove hardcoded stuff from following */
void
set_ps_option(void)
{
    int flag = resource.postscript;

    Widget tmp1 = XtNameToWidget(menus[4], "button_0");
    Widget tmp2 = XtNameToWidget(menus[4], "button_1");
    Widget tmp3 = XtNameToWidget(menus[4], "button_2");

    ASSERT(tmp1 != 0, "Widget not found");
    ASSERT(tmp2 != 0, "Widget not found");
    ASSERT(tmp3 != 0, "Widget not found");

    if (!resource.useGS)
	flag = 0;
    
    XmToggleButtonSetState(tmp1, flag == 1 ? True : False, False);
    XmToggleButtonSetState(tmp2, flag == 2 ? True : False, False);
    XmToggleButtonSetState(tmp3, flag == 0 ? True : False, False);
}

/* FIXME: remove hardcoded stuff from following */
void
set_gs_option(void)
{
    Widget tmp = XtNameToWidget(menus[4], "button_3");
    ASSERT(tmp != 0, "Widget not found");
    XmToggleButtonSetState(tmp, resource.gs_alpha, False);
}

/* FIXME: remove hardcoded stuff from following */
void
set_keep_option(void)
{
    Widget tmp = XtNameToWidget(menus[4], "button_5");
    ASSERT(tmp != 0, "Widget not found");
    XmToggleButtonSetState(tmp, resource.keep_flag, False);
}

/* FIXME: remove hardcoded stuff from following */
void
set_toggle_ruler(void)
{
    Widget tmp = XtNameToWidget(menus[4], "button_4");
    ASSERT(tmp != 0, "Widget not found");
    XmToggleButtonSetState(tmp, resource.ruler_mode, False);
}

/* FIXME: remove hardcoded stuff from following */
void
set_use_tex_option(void)
{
    Widget tmp = XtNameToWidget(menus[4], "button_6");
    ASSERT(tmp != 0, "Widget not found");
    XmToggleButtonSetState(tmp, resource.use_tex_pages, False);
}

/* FIXME: remove hardcoded stuff from following */
void
set_show_statusline_option()
{
    Widget tmp = XtNameToWidget(menus[4], "button_7");
    ASSERT(tmp != 0, "Widget not found");
    XmToggleButtonSetState(tmp, (expert_flag & XPRT_SHOW_STATUSLINE) != 0, False);
}

/* FIXME: remove hardcoded stuff from following */
void
set_show_scrollbars_option(void)
{
    Widget tmp = XtNameToWidget(menus[4], "button_8");
    ASSERT(tmp != 0, "Widget not found");
    XmToggleButtonSetState(tmp, (expert_flag & XPRT_SHOW_SCROLLBARS) != 0, False);
}

/* FIXME: remove hardcoded stuff from following */
void
set_show_pagelist_option(void)
{
    Widget tmp = XtNameToWidget(menus[4], "button_9");
    ASSERT(tmp != 0, "Widget not found");
    XmToggleButtonSetState(tmp, (expert_flag & XPRT_SHOW_PAGELIST) != 0, False);
}

/* FIXME: remove hardcoded stuff from following */
void
set_show_toolbar_option(void)
{
    Widget tmp = XtNameToWidget(menus[4], "button_10");
    ASSERT(tmp != 0, "Widget not found");
    XmToggleButtonSetState(tmp, (expert_flag & XPRT_SHOW_TOOLBAR) != 0, False);
}

void
set_shrink_factor(void)
{
    int i, curr = 0;
    static Widget active_shrink_button = NULL;
    int val = mane.shrinkfactor;
    
    for (i = 0; scale_pulldown_info[i].title != NULL; i++) {
	if (scale_pulldown_info[i].type == XmRADIOBUTTON)
	    curr++;
	if (curr == val)
	    break;
    }
    if (curr > 0) {
	Widget new_shrink_button = NULL;
	char buf[128];
	sprintf(buf, "button_%d", curr - 1);
	new_shrink_button = XtNameToWidget(menus[2], buf); /* FIXME: remove hard-coding of '2' here! */
	/* FIXME: This will toggle the button nevertheless (automatic callback on radiobutton?);
	   but that was different in xdvik-22.48-alpha3. Need to investigate. */
	if (active_shrink_button != NULL)
	    XmToggleButtonSetState(active_shrink_button, False, False);
	if (new_shrink_button != NULL)
	    XmToggleButtonSetState(new_shrink_button, True, False);
	active_shrink_button = new_shrink_button;
    }

    handle_statusline_resize();
#if HAVE_XPM
    tb_set_zoom_sensitivity(val == 1 ? False : True);
#endif
}



/*
 * There's an apparent bug in Motif related to the interaction between
 * the menubar menus and the magnifier.
 *
 * If you click on a menu on the menubar and then click on the drawing
 * widget to pop up a magnifier, the keyboard and pointer are still
 * grabbed, leading to a weird situation in which the magnifier stays
 * around even after you release the pointer.  The ungrab_serial
 * counter works around this bug by ignoring such pointer events.
 *
 * This bug occurs in Motif 1.2 and OpenMotif 2.2.2 (at least).  It
 * does not occur in Lesstif.
 */
static unsigned int ungrab_serial = 0;

static void
popdown_callback(Widget w, XtPointer client_data, XtPointer call_data)
{
    UNUSED(w);
    UNUSED(client_data);
    /* Lesstif gives call_data == NULL */
    if (call_data != NULL && *((XtGrabKind *) call_data) != XtGrabNone)
	ungrab_serial = LastKnownRequestProcessed(DISP);
}

unsigned int
get_last_ungrab(void) {
    return ungrab_serial;
}


void
toggle_menubar(void)
{
    if ((expert_flag & XPRT_SHOW_MENUBAR) == 0)
	XtUnmanageChild(menu_bar);
    else
	XtManageChild(menu_bar);
}    

Widget create_menu_bar(Widget parent)
{
    Widget menu_bar;
    Widget tmp;
    int i;
    char buf[128];
    
    struct menu_item_info menu_info[] = {
	{ "File",	'F',	(XtCallbackProc)file_pulldown_callback,		file_pulldown_info	},
	{ "Navigate",	'N',	(XtCallbackProc)navigate_pulldown_callback,	navigate_pulldown_info	},
	{ "Zoom",	'Z',	(XtCallbackProc)scale_pulldown_callback,	scale_pulldown_info	},
	{ "Mark",	'M',	(XtCallbackProc)mark_pulldown_callback,		mark_pulldown_info	},
	{ "Options",	'O',	(XtCallbackProc)options_pulldown_callback,	options_pulldown_info	},
	{ "Help",	'H',	(XtCallbackProc)help_pulldown_callback,		help_pulldown_info	},
	{ NULL,		0,	NULL,				NULL			},
    };

    /* TODO: create this from the menu_info description as well */
    menu_bar = XmVaCreateSimpleMenuBar(parent, "menuBar",
				       XmVaCASCADEBUTTON, (XtArgVal)XmCvtCTToXmString("File"), 'F',
				       XmVaCASCADEBUTTON, (XtArgVal)XmCvtCTToXmString("Navigate"), 'N',
				       XmVaCASCADEBUTTON, (XtArgVal)XmCvtCTToXmString("Zoom"), 'Z',
				       XmVaCASCADEBUTTON, (XtArgVal)XmCvtCTToXmString("Mark"), 'M',
				       XmVaCASCADEBUTTON, (XtArgVal)XmCvtCTToXmString("Options"), 'O',
				       XmVaCASCADEBUTTON, (XtArgVal)XmCvtCTToXmString("Help"), 'H',
				       NULL, 0);

    for (i = 0; menu_info[i].title != NULL; i++) {
	create_pulldown_menu(i, menu_bar, menu_info[i]);
    }
    
    /*
      workaround for pointer grabbing bug; add additional callback to all menus
      (they have the same parent)
    */
    XtAddCallback(XtParent(menus[i - 1]), XtNpopdownCallback,
		  popdown_callback, (XtPointer) 0);

    sprintf(buf, "button_%d", i - 1);
    if ((tmp = XtNameToWidget(menu_bar, buf)) != NULL)
	XtVaSetValues(menu_bar, XmNmenuHelpWidget, tmp, NULL);

    set_shrink_factor();
    set_keep_option();
    set_use_tex_option();
    set_show_statusline_option();
    set_show_toolbar_option();
    set_show_pagelist_option();
    set_show_scrollbars_option();
    set_ps_option();
    set_gs_option();
    return menu_bar;
}

#else
/* silence `empty compilation unit' warnings */
static void bar(void); static void foo() { bar(); } static void bar(void) { foo(); }
#endif /* MOTIF */
